import importlib
from inspect import getmembers, isfunction
import logging
import os
from os.path import join as pjoin
import shutil
import sys
from tempfile import TemporaryDirectory, mkstemp

import numpy as np
import numpy.testing as npt
import pytest

import dipy.core.gradients as grad
from dipy.data import default_sphere, get_fnames
from dipy.data.fetcher import dipy_home
from dipy.direction.peaks import PeaksAndMetrics
from dipy.io.image import load_nifti, save_nifti
from dipy.io.peaks import load_pam, save_pam
from dipy.io.streamline import load_tractogram
from dipy.io.utils import nifti1_symmat
from dipy.reconst import dti, utils as reconst_utils
from dipy.reconst.shm import convert_sh_descoteaux_tournier
from dipy.testing import assert_true
from dipy.utils.optpkg import optional_package
from dipy.utils.tripwire import TripWireError
from dipy.workflows.io import (
    ConcatenateTractogramFlow,
    ConvertSHFlow,
    ConvertTensorsFlow,
    ConvertTractogramFlow,
    ExtractB0Flow,
    ExtractShellFlow,
    ExtractVolumeFlow,
    FetchFlow,
    IoInfoFlow,
    MathFlow,
    NiftisToPamFlow,
    PamToNiftisFlow,
    SplitFlow,
    TensorToPamFlow,
)

ne, have_ne, _ = optional_package("numexpr")

fname_log = mkstemp()[1]

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)s %(message)s",
    filename=fname_log,
    filemode="w",
)

is_big_endian = "big" in sys.byteorder.lower()


def test_io_info():
    fimg, fbvals, fbvecs = get_fnames(name="small_101D")
    io_info_flow = IoInfoFlow()
    io_info_flow.run([fimg, fbvals, fbvecs])

    fimg, fbvals, fvecs = get_fnames(name="small_25")
    io_info_flow = IoInfoFlow()
    io_info_flow.run([fimg, fbvals, fvecs])

    io_info_flow = IoInfoFlow()
    io_info_flow.run([fimg, fbvals, fvecs], b0_threshold=20, bvecs_tol=0.001)

    filepath_dix, _, _ = get_fnames(name="gold_standard_tracks")
    if not is_big_endian:
        io_info_flow = IoInfoFlow()
        io_info_flow.run(filepath_dix["gs.trx"])

    io_info_flow = IoInfoFlow()
    io_info_flow.run(filepath_dix["gs.trk"])

    io_info_flow = IoInfoFlow()
    npt.assert_raises(TypeError, io_info_flow.run, filepath_dix["gs.tck"])

    io_info_flow = IoInfoFlow()
    io_info_flow.run(filepath_dix["gs.tck"], reference=filepath_dix["gs.nii"])

    with open(fname_log, "r") as file:
        lines = file.readlines()
        try:
            npt.assert_equal(lines[-3], "INFO Total number of unit bvectors 25\n")
        except IndexError:  # logging maybe disabled in IDE setting
            pass


def test_io_fetch():
    fetch_flow = FetchFlow()
    with TemporaryDirectory() as out_dir:
        fetch_flow.run(["bundle_fa_hcp"])
        npt.assert_equal(os.path.isdir(os.path.join(dipy_home, "bundle_fa_hcp")), True)

        fetch_flow.run(["bundle_fa_hcp"], out_dir=out_dir)
        npt.assert_equal(os.path.isdir(os.path.join(out_dir, "bundle_fa_hcp")), True)


def test_io_fetch_fetcher_datanames():
    available_data = FetchFlow.get_fetcher_datanames()

    module_path = "dipy.data.fetcher"
    if module_path in sys.modules:
        fetcher_module = importlib.reload(sys.modules[module_path])
    else:
        fetcher_module = importlib.import_module(module_path)

    ignored_fetchers = ["fetch_data"]
    fetcher_list = {
        name.replace("fetch_", ""): func
        for name, func in getmembers(fetcher_module, isfunction)
        if name.lower().startswith("fetch_") and name.lower() not in ignored_fetchers
    }

    num_expected_fetch_methods = len(fetcher_list)
    npt.assert_equal(len(available_data), num_expected_fetch_methods)
    npt.assert_equal(
        all(dataset_name in available_data.keys() for dataset_name in fetcher_list),
        True,
    )


def test_split_flow():
    with TemporaryDirectory() as out_dir:
        split_flow = SplitFlow()
        data_path, _, _ = get_fnames()
        volume, affine = load_nifti(data_path)
        split_flow.run(data_path, out_dir=out_dir)
        assert_true(os.path.isfile(split_flow.last_generated_outputs["out_split"]))
        split_flow._force_overwrite = True
        split_flow.run(data_path, vol_idx=0, out_dir=out_dir)
        split_path = split_flow.last_generated_outputs["out_split"]
        assert_true(os.path.isfile(split_path))
        split_data, split_affine = load_nifti(split_path)
        npt.assert_equal(split_data.shape, volume[..., 0].shape)
        npt.assert_array_almost_equal(split_affine, affine)


def test_concatenate_flow():
    with TemporaryDirectory() as out_dir:
        concatenate_flow = ConcatenateTractogramFlow()
        data_path, _, _ = get_fnames(name="gold_standard_tracks")
        input_files = [
            v
            for k, v in data_path.items()
            if k in ["gs.trk", "gs.tck", "gs.trx", "gs.fib"]
        ]
        concatenate_flow.run(*input_files, out_dir=out_dir)
        assert_true(
            concatenate_flow.last_generated_outputs["out_extension"].endswith("trx")
        )
        assert_true(
            os.path.isfile(
                concatenate_flow.last_generated_outputs["out_tractogram"] + ".trx"
            )
        )

        trk = load_tractogram(
            concatenate_flow.last_generated_outputs["out_tractogram"] + ".trx", "same"
        )
        npt.assert_equal(len(trk), 13)


def test_convert_sh_flow():
    with TemporaryDirectory() as out_dir:
        filepath_in = os.path.join(out_dir, "sh_coeff_img.nii.gz")
        filename_out = "sh_coeff_img_converted.nii.gz"
        filepath_out = os.path.join(out_dir, filename_out)

        # Create an input image
        dim0, dim1, dim2 = 2, 3, 3  # spatial dimensions of array
        num_sh_coeffs = 15  # 15 sh coeffs means l_max is 4
        img_in = np.arange(dim0 * dim1 * dim2 * num_sh_coeffs, dtype=float).reshape(
            dim0, dim1, dim2, num_sh_coeffs
        )
        save_nifti(filepath_in, img_in, np.eye(4))

        # Compute expected result to compare against later
        expected_img_out = convert_sh_descoteaux_tournier(img_in)

        # Run the workflow and load the output
        workflow = ConvertSHFlow()
        workflow.run(
            filepath_in,
            out_dir=out_dir,
            out_file=filename_out,
        )
        img_out, _ = load_nifti(filepath_out)

        # Compare
        npt.assert_array_almost_equal(img_out, expected_img_out)


def test_convert_tractogram_flow():
    with TemporaryDirectory() as out_dir:
        data_path, _, _ = get_fnames(name="gold_standard_tracks")
        input_files = [
            v
            for k, v in data_path.items()
            if k
            in [
                "gs.tck",
            ]
        ]

        convert_tractogram_flow = ConvertTractogramFlow(mix_names=True)
        convert_tractogram_flow.run(
            input_files, reference=data_path["gs.nii"], out_dir=out_dir
        )

        convert_tractogram_flow._force_overwrite = True
        npt.assert_raises(
            ValueError, convert_tractogram_flow.run, input_files, out_dir=out_dir
        )

        if not is_big_endian:
            npt.assert_warns(
                UserWarning,
                convert_tractogram_flow.run,
                data_path["gs.trx"],
                out_dir=out_dir,
                out_tractogram="gs_converted.trx",
            )


def test_convert_tensors_flow():
    with TemporaryDirectory() as out_dir:
        filepath_in = os.path.join(out_dir, "tensors_img.nii.gz")
        filename_out = "tensors_converted.nii.gz"
        filepath_out = os.path.join(out_dir, filename_out)

        # Create an input image
        fdata, fbval, fbvec = get_fnames(name="small_25")
        data, affine = load_nifti(fdata)
        gtab = grad.gradient_table(fbval, bvecs=fbvec)
        tenmodel = dti.TensorModel(gtab)
        tenfit = tenmodel.fit(data)

        tensor_vals = dti.lower_triangular(tenfit.quadratic_form)
        ten_img = nifti1_symmat(tensor_vals, affine=affine)

        save_nifti(filepath_in, ten_img.get_fdata().squeeze(), affine)

        # Compute expected result to compare against later
        expected_img_out = reconst_utils.convert_tensors(
            ten_img.get_fdata(), "dipy", "mrtrix"
        )

        # Run the workflow and load the output
        workflow = ConvertTensorsFlow()
        workflow.run(
            filepath_in,
            from_format="dipy",
            to_format="mrtrix",
            out_dir=out_dir,
            out_tensor=filename_out,
        )

        img_out, _ = load_nifti(filepath_out)
        npt.assert_array_almost_equal(img_out, expected_img_out)


def generate_random_pam():
    pam = PeaksAndMetrics()
    pam.affine = np.eye(4)
    pam.peak_dirs = np.random.rand(15, 15, 15, 5, 3)
    pam.peak_values = np.zeros((15, 15, 15, 5))
    pam.peak_indices = np.zeros((15, 15, 15, 5))
    pam.shm_coeff = np.zeros((15, 15, 15, 45))
    pam.sphere = default_sphere
    pam.B = np.zeros((45, default_sphere.vertices.shape[0]))
    pam.total_weight = 0.5
    pam.ang_thr = 60
    pam.gfa = np.zeros((10, 10, 10))
    pam.qa = np.zeros((10, 10, 10, 5))
    pam.odf = np.zeros((10, 10, 10, default_sphere.vertices.shape[0]))
    return pam


def test_niftis_to_pam_flow():
    pam = generate_random_pam()
    with TemporaryDirectory() as out_dir:
        fname = pjoin(out_dir, "test.pam5")
        save_pam(fname, pam)

        args = [fname, out_dir]
        flow = PamToNiftisFlow()
        flow.run(*args)

        args = [
            flow.last_generated_outputs["out_peaks_dir"],
            flow.last_generated_outputs["out_peaks_values"],
            flow.last_generated_outputs["out_peaks_indices"],
        ]

        flow2 = NiftisToPamFlow()
        flow2.run(*args, out_dir=out_dir)
        pam_file = flow2.last_generated_outputs["out_pam"]
        assert_true(os.path.isfile(pam_file))

        res_pam = load_pam(pam_file)
        npt.assert_array_equal(pam.affine, res_pam.affine)
        npt.assert_array_almost_equal(pam.peak_dirs, res_pam.peak_dirs)
        npt.assert_array_almost_equal(pam.peak_values, res_pam.peak_values)
        npt.assert_array_almost_equal(pam.peak_indices, res_pam.peak_indices)


def test_tensor_to_pam_flow():
    fdata, fbval, fbvec = get_fnames(name="small_25")
    gtab = grad.gradient_table(fbval, bvecs=fbvec)
    data, affine = load_nifti(fdata)
    dm = dti.TensorModel(gtab)
    df = dm.fit(data)
    df.evals[0, 0, 0] = np.array([0, 0, 0])

    with TemporaryDirectory() as out_dir:
        f_mevals, f_mevecs = (
            pjoin(out_dir, "evals.nii.gz"),
            pjoin(out_dir, "evecs.nii.gz"),
        )
        save_nifti(f_mevals, df.evals, affine)
        save_nifti(f_mevecs, df.evecs, affine)

        args = [f_mevals, f_mevecs]
        flow = TensorToPamFlow()
        flow.run(*args, out_dir=out_dir)
        pam_file = flow.last_generated_outputs["out_pam"]
        assert_true(os.path.isfile(pam_file))

        pam = load_pam(pam_file)
        npt.assert_array_equal(pam.affine, affine)
        npt.assert_array_almost_equal(pam.peak_dirs[..., :3, :], df.evecs)
        npt.assert_array_almost_equal(pam.peak_values[..., :3], df.evals)


def test_pam_to_niftis_flow():
    pam = generate_random_pam()

    with TemporaryDirectory() as out_dir:
        fname = pjoin(out_dir, "test.pam5")
        save_pam(fname, pam)

        args = [fname, out_dir]
        flow = PamToNiftisFlow()
        flow.run(*args)
        assert_true(os.path.isfile(flow.last_generated_outputs["out_peaks_dir"]))
        assert_true(os.path.isfile(flow.last_generated_outputs["out_peaks_values"]))
        assert_true(os.path.isfile(flow.last_generated_outputs["out_peaks_indices"]))
        assert_true(os.path.isfile(flow.last_generated_outputs["out_shm"]))
        assert_true(os.path.isfile(flow.last_generated_outputs["out_gfa"]))
        assert_true(os.path.isfile(flow.last_generated_outputs["out_sphere"]))


def test_math():
    with TemporaryDirectory() as out_dir:
        data_path, _, _ = get_fnames(name="small_101D")
        data_path_a = pjoin(out_dir, "data_a.nii.gz")
        data_path_b = pjoin(out_dir, "data_b.nii.gz")
        shutil.copy(data_path, data_path_a)
        shutil.copy(data_path, data_path_b)

        data, _ = load_nifti(data_path)
        operations = ["vol1*3", "vol1+vol2+vol3", "5*vol1-vol2-vol3", "vol3*2 + vol2"]
        kwargs = [{"dtype": "i"}, {"dtype": "float32"}, {}, {}]

        if have_ne:
            for op, kwarg in zip(operations, kwargs):
                math_flow = MathFlow()
                math_flow.run(
                    op, [data_path_a, data_path_b, data_path], out_dir=out_dir, **kwarg
                )
                out_path = pjoin(out_dir, "math_out.nii.gz")
                out_data, _ = load_nifti(out_path)
                npt.assert_array_equal(out_data, data * 3)
                if kwarg:
                    npt.assert_equal(out_data.dtype, np.dtype(kwarg["dtype"]))

        else:
            math_flow = MathFlow()
            npt.assert_raises(TripWireError, math_flow.run, "vol1*3", [data_path_a])


@pytest.mark.skipif(not have_ne, reason="numexpr not installed")
def test_math_error():
    with TemporaryDirectory() as out_dir:
        data_path, _, _ = get_fnames(name="small_101D")
        data_path_2, _, _ = get_fnames(name="small_64D")
        data_path_a = pjoin(out_dir, "data_a.nii.gz")
        data_path_b = pjoin(out_dir, "data_b.gz")
        data_path_c = pjoin(out_dir, "data_c.nii")
        shutil.copy(data_path, data_path_a)
        shutil.copy(data_path, data_path_b)

        math_flow = MathFlow()
        npt.assert_raises(
            SyntaxError, math_flow.run, "vol1*", [data_path_a], out_dir=out_dir
        )
        npt.assert_raises(
            SystemExit,
            math_flow.run,
            "vol1*2",
            [data_path_a],
            dtype="k",
            out_dir=out_dir,
        )
        npt.assert_raises(
            SystemExit, math_flow.run, "vol1*2", [data_path_b], out_dir=out_dir
        )
        npt.assert_raises(
            SystemExit, math_flow.run, "vol1*2", [data_path_c], out_dir=out_dir
        )
        npt.assert_raises(
            SystemExit, math_flow.run, "vol1*vol3", [data_path_a], out_dir=out_dir
        )
        npt.assert_raises(
            SystemExit,
            math_flow.run,
            "vol1*vol2",
            [data_path, data_path_2],
            out_dir=out_dir,
        )


def test_extract_b0_flow():
    with TemporaryDirectory() as out_dir:
        fdata, fbval, fbvec = get_fnames(name="small_25")
        data, affine = load_nifti(fdata)
        b0_data = data[..., 0]
        b0_path = pjoin(out_dir, "b0_expected.nii.gz")
        save_nifti(b0_path, b0_data, affine)

        extract_b0_flow = ExtractB0Flow()
        extract_b0_flow.run(fdata, fbval, out_dir=out_dir, strategy="first")
        npt.assert_equal(
            extract_b0_flow.last_generated_outputs["out_b0"],
            pjoin(out_dir, "b0.nii.gz"),
        )
        res, _ = load_nifti(extract_b0_flow.last_generated_outputs["out_b0"])
        npt.assert_array_equal(res, b0_data)


def test_extract_shell_flow():
    with TemporaryDirectory() as out_dir:
        fdata, fbval, fbvec = get_fnames(name="small_25")
        data, affine = load_nifti(fdata)

        extract_shell_flow = ExtractShellFlow()
        extract_shell_flow.run(
            fdata, fbval, fbvec, bvals_to_extract="2000", out_dir=out_dir
        )
        res, _ = load_nifti(pjoin(out_dir, "shell_2000.nii.gz"))
        npt.assert_array_equal(res, data[..., 1:])

        extract_shell_flow._force_overwrite = True
        extract_shell_flow.run(
            fdata,
            fbval,
            fbvec,
            bvals_to_extract="0, 2000",
            group_shells=False,
            out_dir=out_dir,
        )
        npt.assert_equal(os.path.isfile(pjoin(out_dir, "shell_0.nii.gz")), True)
        npt.assert_equal(os.path.isfile(pjoin(out_dir, "shell_2000.nii.gz")), True)
        res0, _ = load_nifti(pjoin(out_dir, "shell_0.nii.gz"))
        res2000, _ = load_nifti(pjoin(out_dir, "shell_2000.nii.gz"))
        npt.assert_array_equal(np.squeeze(res0), data[..., 0])
        npt.assert_array_equal(res2000[..., 9], data[..., 10])


def test_extract_volume_flow():
    with TemporaryDirectory() as out_dir:
        fdata, _, _ = get_fnames(name="small_25")
        data, affine = load_nifti(fdata)

        extract_volume_flow = ExtractVolumeFlow()
        extract_volume_flow.run(fdata, vol_idx="0-3,5", out_dir=out_dir)
        res, _ = load_nifti(extract_volume_flow.last_generated_outputs["out_vol"])
        npt.assert_equal(res.shape[-1], 5)

        extract_volume_flow._force_overwrite = True
        extract_volume_flow.run(fdata, vol_idx="0-3,5", grouped=False, out_dir=out_dir)
        npt.assert_equal(os.path.isfile(pjoin(out_dir, "volume_2.nii.gz")), True)
        npt.assert_equal(os.path.isfile(pjoin(out_dir, "volume_5.nii.gz")), True)
        res2, _ = load_nifti(pjoin(out_dir, "volume_2.nii.gz"))
        res5, _ = load_nifti(pjoin(out_dir, "volume_5.nii.gz"))
        npt.assert_array_equal(res2, data[..., 2])
        npt.assert_array_equal(res5, data[..., 5])



import numpy as np
import numpy.testing as npt

from dipy.data import SPHERE_FILES


def test_sphere_dtypes():
    for _, sphere_path in SPHERE_FILES.items():
        sphere_data = np.load(sphere_path)
        npt.assert_equal(sphere_data["vertices"].dtype, np.dtype("<f8"))



from os.path import join as pjoin
from tempfile import TemporaryDirectory

import numpy as np
import numpy.testing as npt
import pytest

from dipy.data import get_fnames
from dipy.io.image import load_nifti_data, save_nifti
from dipy.nn.evac import EVACPlus
from dipy.utils.optpkg import optional_package
from dipy.workflows.nn import BiasFieldCorrectionFlow, EVACPlusFlow

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
torch, have_torch, _ = optional_package("torch", min_version="2.2.0")
have_nn = have_tf or have_torch


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_evac_plus_flow():
    with TemporaryDirectory() as out_dir:
        file_path = get_fnames(name="evac_test_data")

        volume = np.load(file_path)["input"][0]
        temp_affine = np.eye(4)
        temp_path = pjoin(out_dir, "temp.nii.gz")
        save_nifti(temp_path, volume, temp_affine)
        save_masked = True

        evac_flow = EVACPlusFlow()
        evac_flow.run(temp_path, out_dir=out_dir, save_masked=save_masked)

        mask_name = evac_flow.last_generated_outputs["out_mask"]
        masked_name = evac_flow.last_generated_outputs["out_masked"]

        evac = EVACPlus()
        mask = evac.predict(volume, temp_affine)
        masked = volume * mask

        result_mask_data = load_nifti_data(pjoin(out_dir, mask_name))
        npt.assert_array_equal(result_mask_data.astype(np.uint8), mask)

        result_masked_data = load_nifti_data(pjoin(out_dir, masked_name))

        npt.assert_array_equal(result_masked_data, masked)


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_correct_biasfield_flow():
    # Test with T1 data
    if have_nn:
        with TemporaryDirectory() as out_dir:
            file_path = get_fnames(name="deepn4_test_data")

            volume = np.load(file_path[0])["img"]
            temp_affine = np.load(file_path[0])["affine"]
            temp_path = pjoin(out_dir, "temp.nii.gz")
            save_nifti(temp_path, volume, temp_affine)

            bias_flow = BiasFieldCorrectionFlow()
            bias_flow.run(temp_path, out_dir=out_dir)

            corrected_name = bias_flow.last_generated_outputs["out_corrected"]

            corrected_data = load_nifti_data(pjoin(out_dir, corrected_name))
            npt.assert_almost_equal(
                corrected_data.mean(), 119.03902876428222, decimal=4
            )

    # Test with DWI data
    with TemporaryDirectory() as out_dir:
        fdata, fbval, fbvec = get_fnames(name="small_25")
        args = {
            "input_files": fdata,
            "bval": fbval,
            "bvec": fbvec,
            "method": "b0",
            "out_dir": out_dir,
        }
        bias_flow = BiasFieldCorrectionFlow()
        bias_flow.run(**args)

        corrected_name = bias_flow.last_generated_outputs["out_corrected"]

        corrected_data = load_nifti_data(pjoin(out_dir, corrected_name))
        npt.assert_almost_equal(corrected_data.mean(), 0.0384615384615, decimal=5)

    args = {
        "input_files": fdata,
        "bval": fbval,
        "bvec": fbvec,
        "method": "random",
        "out_dir": out_dir,
    }
    npt.assert_raises(SystemExit, bias_flow.run, **args)



"""
=================================================
An introduction to the Probabilistic Tractography
=================================================

Probabilistic fiber tracking is a way of reconstructing white matter
connections using diffusion MR imaging. Like deterministic fiber tracking, the
probabilistic approach follows the trajectory of a possible pathway step by
step starting at a seed, however, unlike deterministic tracking, the tracking
direction at each point along the path is chosen at random from a distribution.
The distribution at each point is different and depends on the observed
diffusion data at that point. The distribution of tracking directions at each
point can be represented as a probability mass function (PMF) if the possible
tracking directions are restricted to discrete numbers of well distributed
points on a sphere.

This example is an extension of the
:ref:`sphx_glr_examples_built_quick_start_tracking_introduction_eudx.py`
example. We'll begin by repeating a few steps from that example, loading the
data and fitting a Constrained Spherical Deconvolution (CSD) model.
"""

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst.csdeconv import ConstrainedSphericalDeconvModel, auto_response_ssst
from dipy.tracking.stopping_criterion import BinaryStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.tracker import probabilistic_tracking
from dipy.tracking.utils import seeds_from_mask
from dipy.viz import actor, colormap, has_fury, window

# Enables/disables interactive visualization
interactive = False

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

seed_mask = labels == 2
seeds = seeds_from_mask(seed_mask, affine, density=2)

white_matter = (labels == 1) | (labels == 2)
sc = BinaryStoppingCriterion(white_matter)

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)
csd_model = ConstrainedSphericalDeconvModel(gtab, response, sh_order_max=6)
csd_fit = csd_model.fit(data, mask=white_matter)


###############################################################################
# The Fiber Orientation Distribution (FOD) of the CSD model estimates the
# distribution of small fiber bundles within each voxel. We can use this
# distribution for probabilistic fiber tracking. One way to do this is to
# represent the FOD using a discrete sphere. This discrete FOD can be used by
# ``probabilistic_tracking`` as a PMF (sf or spherical function) for sampling
# tracking directions.

fod = csd_fit.odf(default_sphere)
streamline_generator = probabilistic_tracking(
    seeds,
    sc,
    affine,
    sf=fod,
    random_seed=1,
    sphere=default_sphere,
    max_angle=20,
    step_size=0.2,
)

streamlines = Streamlines(streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_probabilistic_sf.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(
        scene=scene, out_path="tractogram_probabilistic_sf.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using probabilistic tractography from PMF
#
#
#
# One disadvantage of using a discrete PMF to represent possible tracking
# directions is that it tends to take up a lot of memory (RAM). The size of the
# PMF, the FOD in this case, must be equal to the number of possible tracking
# directions on the hemisphere, and every voxel has a unique PMF. In this case
# the data is ``(81, 106, 76)`` and ``small_sphere`` has 181 directions so the
# FOD is ``(81, 106, 76, 181)``. One way to avoid sampling the PMF and holding
# it in memory is to use directly from the spherical
# harmonic (SH) representation of the FOD. By using this approach, we can also
# use a larger sphere, like ``default_sphere`` which has 362 directions on the
# hemisphere, without having to worry about memory limitations.

streamline_generator = probabilistic_tracking(
    seeds,
    sc,
    affine,
    sh=csd_fit.shm_coeff,
    random_seed=1,
    sphere=default_sphere,
    max_angle=20,
    step_size=0.2,
)

streamlines = Streamlines(streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_probabilistic_sh.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(
        scene=scene, out_path="tractogram_probabilistic_sh.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using probabilistic tractography from SH
#
#
#
# Not all model fits have the ``shm_coeff`` attribute because not all models
# use this basis to represent the data internally. However we can fit the ODF
# of any model to the spherical harmonic basis using the ``peaks_from_model``
# function.

peaks = peaks_from_model(
    csd_model,
    data,
    default_sphere,
    0.5,
    25,
    mask=white_matter,
    return_sh=True,
    parallel=True,
    num_processes=1,
)
fod_coeff = peaks.shm_coeff

streamline_generator = probabilistic_tracking(
    seeds,
    sc,
    affine,
    sh=fod_coeff,
    random_seed=1,
    sphere=default_sphere,
    max_angle=20,
    step_size=0.2,
)

streamlines = Streamlines(streamline_generator)
sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_probabilistic_sh_pfm.trk")

if has_fury:
    scene = window.Scene()
    scene.add(actor.line(streamlines, colors=colormap.line_colors(streamlines)))
    window.record(
        scene=scene, out_path="tractogram_probabilistic_sh_pfm.png", size=(800, 800)
    )
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Corpus Callosum using probabilistic tractography from SH
# (peaks_from_model)



"""
=========================
Calculate Path Length Map
=========================

We show how to calculate a Path Length Map for Anisotropic Radiation Therapy
Contours given a set of streamlines and a region of interest (ROI).
The Path Length Map is a volume in which each voxel's value is the shortest
distance along a streamline to a given region of interest (ROI). This map can
be used to anisotropically modify radiation therapy treatment contours based
on a tractography model of the local white matter anatomy, as described in
:footcite:p:`Jordan2019`, by executing this tutorial with the gross tumor volume
(GTV) as the ROI.

.. note::

    The background value is set to -1 by default


Let's start by importing the necessary modules.
"""

import matplotlib as mpl
from mpl_toolkits.axes_grid1 import AxesGrid
import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.reconst.shm import CsaOdfModel
from dipy.tracking import utils
from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.tracking.utils import path_length
from dipy.viz import actor, colormap as cmap, window

###############################################################################
# First, we need to generate some streamlines and visualize. For a more
# complete description of these steps, please refer to the
# :ref:`sphx_glr_examples_built_fiber_tracking_tracking_probabilistic.py`
# and the Visualization of ROI Surface Rendered with Streamlines Tutorials.

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

white_matter = (labels == 1) | (labels == 2)

csa_model = CsaOdfModel(gtab, sh_order_max=6)
csa_peaks = peaks_from_model(
    csa_model,
    data,
    default_sphere,
    relative_peak_threshold=0.8,
    min_separation_angle=45,
    mask=white_matter,
)

stopping_criterion = ThresholdStoppingCriterion(csa_peaks.gfa, 0.25)

###############################################################################
# We will use an anatomically-based corpus callosum ROI as our seed mask to
# demonstrate the method. In practice, this corpus callosum mask (labels == 2)
# should be replaced with the desired ROI mask (e.g. gross tumor volume (GTV),
# lesion mask, or electrode mask).

# Make a corpus callosum seed mask for tracking
seed_mask = labels == 2
seeds = utils.seeds_from_mask(seed_mask, affine, density=[1, 1, 1])

# Make a streamline bundle model of the corpus callosum ROI connectivity
streamlines = LocalTracking(csa_peaks, stopping_criterion, seeds, affine, step_size=2)
streamlines = Streamlines(streamlines)

# Visualize the streamlines and the Path Length Map base ROI
# (in this case also the seed ROI)

streamlines_actor = actor.line(streamlines, colors=cmap.line_colors(streamlines))
surface_opacity = 0.5
surface_color = [0, 1, 1]
seedroi_actor = actor.contour_from_roi(
    seed_mask, affine=affine, color=surface_color, opacity=surface_opacity
)

scene = window.Scene()
scene.add(streamlines_actor)
scene.add(seedroi_actor)

###############################################################################
# If you set interactive to True (below), the scene will pop up in an
# interactive window.

interactive = False
if interactive:
    window.show(scene)

window.record(scene=scene, n_frames=1, out_path="plm_roi_sls.png", size=(800, 800))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# A top view of corpus callosum streamlines with the blue transparent ROI in
# the center.
#
#
#
# Now we calculate the Path Length Map using the corpus callosum streamline
# bundle and corpus callosum ROI.
#
# **NOTE**: the mask used to seed the tracking does not have to be the Path
# Length Map base ROI, as we do here, but it often makes sense for them to be
# the same ROI if we want a map of the whole brain's distance back to our ROI.
# (e.g. we could test a hypothesis about the motor system by making a
# streamline bundle model of the cortico-spinal track (CST) and input a
# lesion mask as our Path Length Map base ROI to restrict the analysis to
# the CST)

path_length_map_base_roi = seed_mask

# calculate the WMPL
wmpl = path_length(streamlines, affine, path_length_map_base_roi)

# save the WMPL as a nifti
save_nifti("example_cc_path_length_map.nii.gz", wmpl.astype(np.float32), affine)

# get the T1 to show anatomical context of the WMPL
t1_fname = get_fnames(name="stanford_t1")
t1_data = load_nifti_data(t1_fname)


fig = mpl.pyplot.figure()
fig.subplots_adjust(left=0.05, right=0.95)
ax = AxesGrid(
    fig,
    111,
    nrows_ncols=(1, 3),
    cbar_location="right",
    cbar_mode="single",
    cbar_size="10%",
    cbar_pad="5%",
)

###############################################################################
# We will mask our WMPL to ignore values less than zero because negative
# numbers indicate no path back to the ROI was found in the provided
# streamlines

wmpl_show = np.ma.masked_where(wmpl < 0, wmpl)

slx, sly, slz = [60, 50, 35]
ax[0].matshow(np.rot90(t1_data[:, slx, :]), cmap=mpl.cm.bone)
im = ax[0].matshow(np.rot90(wmpl_show[:, slx, :]), cmap=mpl.cm.cool, vmin=0, vmax=80)

ax[1].matshow(np.rot90(t1_data[:, sly, :]), cmap=mpl.cm.bone)
im = ax[1].matshow(np.rot90(wmpl_show[:, sly, :]), cmap=mpl.cm.cool, vmin=0, vmax=80)

ax[2].matshow(np.rot90(t1_data[:, slz, :]), cmap=mpl.cm.bone)
im = ax[2].matshow(np.rot90(wmpl_show[:, slz, :]), cmap=mpl.cm.cool, vmin=0, vmax=80)

ax.cbar_axes[0].colorbar(im)
for lax in ax:
    lax.set_xticks([])
    lax.set_yticks([])
fig.savefig("Path_Length_Map.png")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Path Length Map showing the shortest distance, along a streamline,
# from the corpus callosum ROI with the background set to -1.
#
# References
# ----------
#
# .. footbibliography::
#



"""
=======================================
Tracking with the Sparse Fascicle Model
=======================================

Tracking requires a per-voxel model. Here, the model is the Sparse Fascicle
Model (SFM), described in :footcite:p:`Rokem2015`. This model reconstructs the
diffusion signal as a combination of the signals from different fascicles (see
also :ref:`sphx_glr_examples_built_reconstruction_reconst_sfm.py`).
"""

from numpy.linalg import inv

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, get_sphere
from dipy.direction.peaks import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import save_trk
from dipy.reconst import sfm
from dipy.reconst.csdeconv import auto_response_ssst
from dipy.tracking import utils
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import (
    Streamlines,
    select_random_set_of_streamlines,
    transform_streamlines,
)
from dipy.tracking.tracker import eudx_tracking
from dipy.viz import actor, colormap, has_fury, window

# Enables/disables interactive visualization
interactive = False

###############################################################################
# To begin, we read the Stanford HARDI data set into memory:

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

###############################################################################
# This data set provides a label map (generated using FreeSurfer_, in which the
# white matter voxels are labeled as either 1 or 2:

white_matter = (labels == 1) | (labels == 2)

###############################################################################
# The first step in tracking is generating a model from which tracking
# directions can be extracted in every voxel.

# For the SFM, this requires first that we define a canonical response function
# that will be used to deconvolve the signal in every voxel

response, ratio = auto_response_ssst(gtab, data, roi_radii=10, fa_thr=0.7)

###############################################################################
# We initialize an SFM model object, using this response function and using
# the default sphere (362  vertices, symmetrically distributed on the surface
# of the sphere):

sphere = get_sphere()
sf_model = sfm.SparseFascicleModel(
    gtab, sphere=sphere, l1_ratio=0.5, alpha=0.001, response=response[0]
)

###############################################################################
# We fit this model to the data in each voxel in the white-matter mask, so that
# we can use these directions in tracking:

pnm = peaks_from_model(
    sf_model,
    data,
    sphere,
    relative_peak_threshold=0.5,
    min_separation_angle=25,
    mask=white_matter,
    parallel=True,
    num_processes=1,
)

###############################################################################
# A ThresholdStoppingCriterion object is used to segment the data to track only
# through areas in which the Generalized Fractional Anisotropy (GFA) is
# sufficiently high.

stopping_criterion = ThresholdStoppingCriterion(pnm.gfa, 0.25)

###############################################################################
# Tracking will be started from a set of seeds evenly distributed in the white
# matter:

seeds = utils.seeds_from_mask(white_matter, affine, density=[2, 2, 2])

###############################################################################
# For the sake of brevity, we will take only the first 1000 seeds, generating
# only 1000 streamlines. Remove this line to track from many more points in
# all of the white matter

seeds = seeds[:1000]

###############################################################################
# We now have the necessary components to construct a tracking pipeline and
# execute the tracking

streamline_generator = eudx_tracking(
    seeds, stopping_criterion, affine, pam=pnm, step_size=0.5
)
streamlines = Streamlines(streamline_generator)

###############################################################################
# Next, we will create a visualization of these streamlines, relative to this
# subject's T1-weighted anatomy:

t1_fname = get_fnames(name="stanford_t1")
t1_data, t1_aff = load_nifti(t1_fname)
color = colormap.line_colors(streamlines)

###############################################################################
# To speed up visualization, we will select a random sub-set of streamlines to
# display. This is particularly important, if you track from seeds throughout
# the entire white matter, generating many streamlines. In this case, for
# demonstration purposes, we select a subset of 900 streamlines.

plot_streamlines = select_random_set_of_streamlines(streamlines, 900)

if has_fury:
    streamlines_actor = actor.streamtube(
        list(transform_streamlines(plot_streamlines, inv(t1_aff))),
        colors=colormap.line_colors(streamlines),
        linewidth=0.1,
    )

    vol_actor = actor.slicer(t1_data)

    vol_actor.display(x=40)
    vol_actor2 = vol_actor.copy()
    vol_actor2.display(z=35)

    scene = window.Scene()
    scene.add(streamlines_actor)
    scene.add(vol_actor)
    scene.add(vol_actor2)

    window.record(scene=scene, out_path="tractogram_sfm.png", size=(800, 800))
    if interactive:
        window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Sparse Fascicle Model tracks
#
#
# Finally, we can save these streamlines to a 'trk' file, for use in other
# software, or for further analysis.

sft = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
save_trk(sft, "tractogram_sfm_detr.trk")

###############################################################################
# References
# ----------
#
# .. footbibliography::
#



import importlib
import sys
import warnings

import numpy as np
from numpy.testing import assert_almost_equal
import pytest

from dipy.data import get_fnames
from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
torch, have_torch, _ = optional_package("torch", min_version="2.2.0")
have_nn = have_tf or have_torch
BACKENDS = [
    backend
    for backend, available in [("tensorflow", have_tf), ("torch", have_torch)]
    if available
]


@pytest.mark.skipif(not have_nn, reason="Requires TensorFlow or Torch")
def test_default_weights(monkeypatch):
    file_names = get_fnames(name="deepn4_test_data")
    input_arr = np.load(file_names[0])["img"]
    input_affine_arr = np.load(file_names[0])["affine"]
    target_arr = np.load(file_names[1])["corr"]

    for backend in BACKENDS:
        monkeypatch.setenv("DIPY_NN_BACKEND", backend)
        with warnings.catch_warnings():
            msg = ".*uses TensorFlow.*install PyTorch.*"
            warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
            dipy_nn = importlib.reload(sys.modules["dipy.nn"])
            deepn4_mod = dipy_nn.deepn4

        deepn4_model = deepn4_mod.DeepN4()
        deepn4_model.fetch_default_weights()
        results_arr = deepn4_model.predict(input_arr, input_affine_arr)
        assert_almost_equal(results_arr / 100, target_arr / 100, decimal=1)



import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_less,
)

from dipy.data import get_sphere
from dipy.reconst.bingham import (
    _bingham_fit_peak,
    _convert_bingham_pars,
    _single_bingham_to_sf,
    _single_sf_to_bingham,
    bingham_fiber_density,
    bingham_fiber_spread,
    k2odi,
    odi2k,
    sf_to_bingham,
    sh_to_bingham,
)
from dipy.reconst.shm import sf_to_sh


def setup_module():
    global sphere
    sphere = get_sphere(name="repulsion724")
    sphere = sphere.subdivide(n=2)


def teardown_module():
    global sphere
    sphere = None


def test_bingham_fit():
    """Tests for bingham function and single Bingham fit"""
    peak_dir = np.array([1, 0, 0])
    ma_axis = np.array([0, 1, 0])
    mi_axis = np.array([0, 0, 1])
    k1 = 2
    k2 = 6
    f0 = 3

    # Test if maximum amplitude is in the expected Bingham main direction
    # which should be perpendicular to both ma_axis and mi_axis
    odf_test = _single_bingham_to_sf(f0, k1, k2, ma_axis, mi_axis, peak_dir)
    assert_almost_equal(odf_test, f0)

    # Test Bingham fit on full sampled GT Bingham function
    odf_gt = _single_bingham_to_sf(f0, k1, k2, ma_axis, mi_axis, sphere.vertices)
    a0, c1, c2, mu0, mu1, mu2 = _bingham_fit_peak(odf_gt, peak_dir, sphere, 45)

    # check scalar parameters
    assert_almost_equal(a0, f0, decimal=3)
    assert_almost_equal(c1, k1, decimal=3)
    assert_almost_equal(c2, k2, decimal=3)

    # check if measured peak direction and dispersion axis are aligned to their
    # respective GT
    Mus = np.array([mu0, mu1, mu2])
    Mus_ref = np.array([peak_dir, ma_axis, mi_axis])
    assert_array_almost_equal(
        np.abs(np.diag(np.dot(Mus, Mus_ref))), np.ones(3), decimal=5
    )

    # check the same for bingham_fit_odf
    fits, n = _single_sf_to_bingham(odf_gt, sphere, max_search_angle=45)
    assert_almost_equal(fits[0][0], f0, decimal=3)
    assert_almost_equal(fits[0][1], k1, decimal=3)
    assert_almost_equal(fits[0][2], k2, decimal=3)
    Mus = np.array([fits[0][3], fits[0][4], fits[0][5]])
    # I had to decrease the precision in the assert below because main peak
    # direction is now calculated (before the GT direction was given)
    assert_array_almost_equal(
        np.abs(np.diag(np.dot(Mus, Mus_ref))), np.ones(3), decimal=5
    )


def test_bingham_metrics():
    axis0 = np.array([1, 0, 0])
    axis1 = np.array([0, 1, 0])
    axis2 = np.array([0, 0, 1])
    k1 = 2
    k2 = 6
    f0_lobe1 = 3
    f0_lobe2 = 1

    # define the parameters of two bingham functions with different amplitudes
    fits = [(f0_lobe1, k1, k2, axis0, axis1, axis2)]
    fits.append((f0_lobe2, k1, k2, axis0, axis1, axis2))

    # First test just to check right parameter conversion
    ref_pars = np.zeros((2, 12))
    ref_pars[0, 0] = f0_lobe1
    ref_pars[1, 0] = f0_lobe2
    ref_pars[0, 1] = ref_pars[1, 1] = k1
    ref_pars[0, 2] = ref_pars[1, 2] = k2
    ref_pars[0, 3:6] = ref_pars[1, 3:6] = axis0
    ref_pars[0, 6:9] = ref_pars[1, 6:9] = axis1
    ref_pars[0, 9:12] = ref_pars[1, 9:12] = axis2
    bpars = _convert_bingham_pars(fits, 2)
    assert_array_almost_equal(bpars, ref_pars)

    # TEST: Bingham Fiber density
    # As the amplitude of the first bingham function is 3 times higher than
    # the second, its surface integral have to be also 3 times larger.
    fd = bingham_fiber_density(bpars)

    assert_almost_equal(fd[0] / fd[1], 3)

    # Fiber density using the default sphere should be close to the fd obtained
    # using a high-resolution sphere (2621442 vertices)
    fd_hires = bingham_fiber_density(bpars, subdivide=9)

    # We must lower the precision for the test to pass, but still shows that
    # the fiber density is precise up to 4 decimals using only 0.39% of the
    # samples (10242 versus 2621442 vertices)
    assert_array_almost_equal(fd, fd_hires, decimal=4)

    # Rotating the Bingham distribution should not bias the FD estimate
    xfits = [
        (f0_lobe1, k1, k2, axis0, axis1, axis2),
        (f0_lobe1, k1, k2, axis1, axis2, axis0),
        (f0_lobe1, k1, k2, axis2, axis0, axis1),
    ]
    xpars = _convert_bingham_pars(xfits, 3)
    xfd = bingham_fiber_density(xpars)

    assert_almost_equal(xfd[0], xfd[1])
    assert_almost_equal(xfd[0], xfd[2])

    # If the Bingham function is a sphere of unit radius, the
    # fiber density should be 4*np.pi.
    sphere_fit = [(1.0, 0.0, 0.0, axis0, axis1, axis2)]
    sphere_pars = _convert_bingham_pars(sphere_fit, 1)

    fd_sphere = bingham_fiber_density(sphere_pars)

    assert_almost_equal(fd_sphere[0], 4.0 * np.pi)

    # TEST: k2odi and odi2k conversions
    assert_almost_equal(odi2k(k2odi(np.array(k1))), k1)
    assert_almost_equal(odi2k(k2odi(np.array(k2))), k2)

    # TEST: Fiber spread
    f0s = np.array([f0_lobe1, f0_lobe2])
    fs = bingham_fiber_spread(f0s, fd)

    assert_array_almost_equal(fs, fd / f0s)


def test_bingham_from_odf():
    # Reconstruct multi voxel ODFs to test bingham_from_odf
    ma_axis = np.array([0, 1, 0])
    mi_axis = np.array([0, 0, 1])
    k1 = 2
    k2 = 6
    f0 = 3
    odf = _single_bingham_to_sf(f0, k1, k2, ma_axis, mi_axis, sphere.vertices)

    # Perform Bingham fit in multi-voxel odf
    multi_odfs = np.zeros((2, 2, 1, len(sphere.vertices)))
    multi_odfs[...] = odf
    bim = sf_to_bingham(multi_odfs, sphere, npeaks=2, max_search_angle=45)

    # check model_params
    assert_almost_equal(bim.model_params[0, 0, 0, 0, 0], f0, decimal=3)
    assert_almost_equal(bim.model_params[0, 0, 0, 0, 1], k1, decimal=3)
    assert_almost_equal(bim.model_params[0, 0, 0, 0, 2], k2, decimal=3)
    # check if estimates for a second lobe are zero (note that a single peak
    # ODF is assumed here for this test GT)
    assert_array_almost_equal(bim.model_params[0, 0, 0, 1], np.zeros(12))

    # check if we have estimates in the right lobe for all voxels
    peak_v = bim.model_params[0, 0, 0, 0, 0]
    assert_array_almost_equal(bim.amplitude_lobe[..., 0], peak_v * np.ones((2, 2, 1)))
    assert_array_almost_equal(bim.amplitude_lobe[..., 1], np.zeros((2, 2, 1)))

    # check kappas
    assert_almost_equal(bim.kappa1_lobe[0, 0, 0, 0], k1, decimal=3)
    assert_almost_equal(bim.kappa2_lobe[0, 0, 0, 0], k2, decimal=3)
    assert_almost_equal(bim.kappa_total_lobe[0, 0, 0, 0], np.sqrt(k1 * k2), decimal=3)

    # check ODI
    assert_almost_equal(bim.odi1_lobe[0, 0, 0, 0], k2odi(np.array(k1)), decimal=3)
    assert_almost_equal(bim.odi2_lobe[0, 0, 0, 0], k2odi(np.array(k2)), decimal=3)
    # ODI2 < ODI total < ODI1
    assert_array_less(bim.odi2_lobe[..., 0], bim.odi1_lobe[..., 0])
    assert_array_less(bim.odi2_lobe[..., 0], bim.odi_total_lobe[..., 0])
    assert_array_less(bim.odi_total_lobe[..., 0], bim.odi1_lobe[..., 0])

    # check fiber_density estimates (larger than zero for lobe 0)
    assert_array_less(np.zeros((2, 2, 1)), bim.fd_lobe[:, :, :, 0])
    assert_almost_equal(np.zeros((2, 2, 1)), bim.fd_lobe[:, :, :, 1])

    # check global metrics: since this simulations only have one lobe, global
    # metrics have to give the same values than their counterparts for lobe 1
    assert_almost_equal(bim.odi1_voxel, bim.odi1_lobe[..., 0])
    assert_almost_equal(bim.odi2_voxel, bim.odi2_lobe[..., 0])
    assert_almost_equal(bim.odi_total_voxel, bim.odi_total_lobe[..., 0])
    assert_almost_equal(bim.fd_voxel, bim.fd_lobe[..., 0])

    # check fiber spread
    fs_v = bim.fd_lobe[0, 0, 0, 0] / peak_v
    assert_almost_equal(bim.fs_lobe[..., 0], fs_v)

    # check reconstructed odf
    reconst_odf = bim.odf(sphere)
    assert_almost_equal(reconst_odf[0, 0, 0], odf, decimal=2)


def test_bingham_from_sh():
    ma_axis = np.array([0, 1, 0])
    mi_axis = np.array([0, 0, 1])
    k1 = 2
    k2 = 6
    f0 = 3
    odf = _single_bingham_to_sf(f0, k1, k2, ma_axis, mi_axis, sphere.vertices)

    bim_odf = sf_to_bingham(odf, sphere, npeaks=2, max_search_angle=45)
    sh = sf_to_sh(odf, sphere, sh_order_max=16, legacy=False)
    bim_sh = sh_to_bingham(sh, sphere, legacy=False, npeaks=2, max_search_angle=45)
    assert_array_almost_equal(bim_sh.model_params, bim_odf.model_params, decimal=3)



from dipy.direction.pmf cimport PmfGen
from dipy.utils.fast_numpy cimport RNGState


cpdef enum TrackerStatus:
    SUCCESS = 1
    FAIL = -1

ctypedef TrackerStatus (*func_ptr)(double* point,
                                   double* direction,
                                   TrackerParameters params,
                                   double* stream_data,
                                   PmfGen pmf_gen,
                                   RNGState* rng) noexcept nogil

cdef class ParallelTransportTrackerParameters:
    cdef public double angular_separation
    cdef public double data_support_exponent
    cdef public double k_small
    cdef public int probe_count
    cdef public double probe_length
    cdef public double probe_normalizer
    cdef public int probe_quality
    cdef public double probe_radius
    cdef public double probe_step_size
    cdef public int rejection_sampling_max_try
    cdef public int rejection_sampling_nbr_sample

cdef class ShTrackerParameters:
    cdef public double pmf_threshold

cdef class TrackerParameters:
    cdef func_ptr tracker

    cdef public double cos_similarity
    cdef public double max_angle
    cdef public double max_curvature
    cdef public int max_len
    cdef public int min_len
    cdef public int random_seed
    cdef public double step_size
    cdef public double average_voxel_size
    cdef public double[3] voxel_size
    cdef public double[3] inv_voxel_size
    cdef public bint return_all

    cdef public ShTrackerParameters sh
    cdef public ParallelTransportTrackerParameters ptt

    cdef void set_tracker_c(self, func_ptr tracker) noexcept nogil



import logging
import os
import sys
from time import time

import numpy as np

from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti
from dipy.io.stateful_tractogram import Space, StatefulTractogram
from dipy.io.streamline import load_tractogram, save_tractogram
from dipy.segment.bundles import RecoBundles
from dipy.segment.mask import median_otsu
from dipy.segment.tissue import TissueClassifierHMRF, dam_classifier
from dipy.tracking import Streamlines
from dipy.workflows.utils import handle_vol_idx
from dipy.workflows.workflow import Workflow


class MedianOtsuFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "medotsu"

    def run(
        self,
        input_files,
        save_masked=False,
        median_radius=2,
        numpass=5,
        autocrop=False,
        vol_idx=None,
        dilate=None,
        finalize_mask=False,
        out_dir="",
        out_mask="brain_mask.nii.gz",
        out_masked="dwi_masked.nii.gz",
    ):
        """Workflow wrapping the median_otsu segmentation method.

        Applies median_otsu segmentation on each file found by 'globing'
        ``input_files`` and saves the results in a directory specified by
        ``out_dir``.

        Parameters
        ----------
        input_files : string
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        save_masked : bool, optional
            Save mask.
        median_radius : int, optional
            Radius (in voxels) of the applied median filter.
        numpass : int, optional
            Number of pass of the median filter.
        autocrop : bool, optional
            If True, the masked input_volumes will also be cropped using the
            bounding box defined by the masked data. For example, if diffusion
            images are of 1x1x1 (mm^3) or higher resolution auto-cropping could
            reduce their size in memory and speed up some of the analysis.
        vol_idx : str, optional
            1D array representing indices of ``axis=-1`` of a 4D
            `input_volume`. From the command line use something like
            '1,2,3-5,7'. This input is required for 4D volumes.
        dilate : int, optional
            number of iterations for binary dilation.
        finalize_mask : bool, optional
            Whether to remove potential holes or islands.
            Useful for solving minor errors.
        out_dir : string, optional
            Output directory.
        out_mask : string, optional
            Name of the mask volume to be saved.
        out_masked : string, optional
            Name of the masked volume to be saved.
        """
        io_it = self.get_io_iterator()
        vol_idx = handle_vol_idx(vol_idx)

        for fpath, mask_out_path, masked_out_path in io_it:
            logging.info(f"Applying median_otsu segmentation on {fpath}")

            data, affine, img = load_nifti(fpath, return_img=True)
            masked_volume, mask_volume = median_otsu(
                data,
                vol_idx=vol_idx,
                median_radius=median_radius,
                numpass=numpass,
                autocrop=autocrop,
                dilate=dilate,
                finalize_mask=finalize_mask,
            )

            save_nifti(mask_out_path, mask_volume.astype(np.float64), affine)

            logging.info(f"Mask saved as {mask_out_path}")

            if save_masked:
                save_nifti(masked_out_path, masked_volume, affine, hdr=img.header)

                logging.info(f"Masked volume saved as {masked_out_path}")

        return io_it


class RecoBundlesFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "recobundles"

    def run(
        self,
        streamline_files,
        model_bundle_files,
        greater_than=50,
        less_than=1000000,
        no_slr=False,
        clust_thr=15.0,
        reduction_thr=15.0,
        reduction_distance="mdf",
        model_clust_thr=2.5,
        pruning_thr=8.0,
        pruning_distance="mdf",
        slr_metric="symmetric",
        slr_transform="similarity",
        slr_matrix="small",
        refine=False,
        r_reduction_thr=12.0,
        r_pruning_thr=6.0,
        no_r_slr=False,
        out_dir="",
        out_recognized_transf="recognized.trk",
        out_recognized_labels="labels.npy",
    ):
        """Recognize bundles

        See :footcite:p:`Garyfallidis2018` and :footcite:p:`Chandio2020a` for
        further details about the method.

        Parameters
        ----------
        streamline_files : string
            The path of streamline files where you want to recognize bundles.
        model_bundle_files : string
            The path of model bundle files.
        greater_than : int, optional
            Keep streamlines that have length greater than
            this value in mm.
        less_than : int, optional
            Keep streamlines have length less than this value
            in mm.
        no_slr : bool, optional
            Don't enable local Streamline-based Linear
            Registration.
        clust_thr : float, optional
            MDF distance threshold for all streamlines.
        reduction_thr : float, optional
            Reduce search space by (mm).
        reduction_distance : string, optional
            Reduction distance type can be mdf or mam.
        model_clust_thr : float, optional
            MDF distance threshold for the model bundles.
        pruning_thr : float, optional
            Pruning after matching.
        pruning_distance : string, optional
            Pruning distance type can be mdf or mam.
        slr_metric : string, optional
            Options are None, symmetric, asymmetric or diagonal.
        slr_transform : string, optional
            Transformation allowed. translation, rigid, similarity or scaling.
        slr_matrix : string, optional
            Options are 'nano', 'tiny', 'small', 'medium', 'large', 'huge'.
        refine : bool, optional
            Enable refine recognized bundle.
        r_reduction_thr : float, optional
            Refine reduce search space by (mm).
        r_pruning_thr : float, optional
            Refine pruning after matching.
        no_r_slr : bool, optional
            Don't enable Refine local Streamline-based Linear
            Registration.
        out_dir : string, optional
            Output directory.
        out_recognized_transf : string, optional
            Recognized bundle in the space of the model bundle.
        out_recognized_labels : string, optional
            Indices of recognized bundle in the original tractogram.

        References
        ----------
        .. footbibliography::

        """
        slr = not no_slr
        r_slr = not no_r_slr

        bounds = [
            (-30, 30),
            (-30, 30),
            (-30, 30),
            (-45, 45),
            (-45, 45),
            (-45, 45),
            (0.8, 1.2),
            (0.8, 1.2),
            (0.8, 1.2),
        ]

        slr_matrix = slr_matrix.lower()
        if slr_matrix == "nano":
            slr_select = (100, 100)
        if slr_matrix == "tiny":
            slr_select = (250, 250)
        if slr_matrix == "small":
            slr_select = (400, 400)
        if slr_matrix == "medium":
            slr_select = (600, 600)
        if slr_matrix == "large":
            slr_select = (800, 800)
        if slr_matrix == "huge":
            slr_select = (1200, 1200)

        slr_transform = slr_transform.lower()
        if slr_transform == "translation":
            bounds = bounds[:3]
        if slr_transform == "rigid":
            bounds = bounds[:6]
        if slr_transform == "similarity":
            bounds = bounds[:7]
        if slr_transform == "scaling":
            bounds = bounds[:9]

        logging.info("### RecoBundles ###")

        io_it = self.get_io_iterator()

        t = time()
        logging.info(streamline_files)
        input_obj = load_tractogram(streamline_files, "same", bbox_valid_check=False)
        streamlines = input_obj.streamlines

        logging.info(f" Loading time {time() - t:0.3f} sec")

        rb = RecoBundles(streamlines, greater_than=greater_than, less_than=less_than)

        for _, mb, out_rec, out_labels in io_it:
            t = time()
            logging.info(mb)
            model_bundle = load_tractogram(
                mb, "same", bbox_valid_check=False
            ).streamlines
            logging.info(f" Loading time {time() - t:0.3f} sec")
            logging.info("model file = ")
            logging.info(mb)

            recognized_bundle, labels = rb.recognize(
                model_bundle,
                model_clust_thr=model_clust_thr,
                reduction_thr=reduction_thr,
                reduction_distance=reduction_distance,
                pruning_thr=pruning_thr,
                pruning_distance=pruning_distance,
                slr=slr,
                slr_metric=slr_metric,
                slr_x0=slr_transform,
                slr_bounds=bounds,
                slr_select=slr_select,
                slr_method="L-BFGS-B",
            )

            if refine:
                if len(recognized_bundle) > 1:
                    # affine
                    x0 = np.array([0, 0, 0, 0, 0, 0, 1.0, 1.0, 1, 0, 0, 0])
                    affine_bounds = [
                        (-30, 30),
                        (-30, 30),
                        (-30, 30),
                        (-45, 45),
                        (-45, 45),
                        (-45, 45),
                        (0.8, 1.2),
                        (0.8, 1.2),
                        (0.8, 1.2),
                        (-10, 10),
                        (-10, 10),
                        (-10, 10),
                    ]

                    recognized_bundle, labels = rb.refine(
                        model_bundle,
                        recognized_bundle,
                        model_clust_thr=model_clust_thr,
                        reduction_thr=r_reduction_thr,
                        reduction_distance=reduction_distance,
                        pruning_thr=r_pruning_thr,
                        pruning_distance=pruning_distance,
                        slr=r_slr,
                        slr_metric=slr_metric,
                        slr_x0=x0,
                        slr_bounds=affine_bounds,
                        slr_select=slr_select,
                        slr_method="L-BFGS-B",
                    )

            if len(labels) > 0:
                ba, bmd = rb.evaluate_results(
                    model_bundle, recognized_bundle, slr_select
                )

                logging.info(f"Bundle adjacency Metric {ba}")
                logging.info(f"Bundle Min Distance Metric {bmd}")

            new_tractogram = StatefulTractogram(
                recognized_bundle, streamline_files, Space.RASMM
            )
            save_tractogram(new_tractogram, out_rec, bbox_valid_check=False)
            logging.info("Saving output files ...")
            np.save(out_labels, np.array(labels))
            logging.info(out_rec)
            logging.info(out_labels)


class LabelsBundlesFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "labelsbundles"

    def run(
        self,
        streamline_files,
        labels_files,
        out_dir="",
        out_bundle="recognized_orig.trk",
    ):
        """Extract bundles using existing indices (labels)

        See :footcite:p:`Garyfallidis2018` for further details about the method.

        Parameters
        ----------
        streamline_files : string
            The path of streamline files where you want to recognize bundles.
        labels_files : string
            The path of model bundle files.
        out_dir : string, optional
            Output directory.
        out_bundle : string, optional
            Recognized bundle in the space of the model bundle.

        References
        ----------
        .. footbibliography::

        """
        logging.info("### Labels to Bundles ###")

        io_it = self.get_io_iterator()
        for f_steamlines, f_labels, out_bundle in io_it:
            logging.info(f_steamlines)
            sft = load_tractogram(f_steamlines, "same", bbox_valid_check=False)
            streamlines = sft.streamlines

            logging.info(f_labels)
            location = np.load(f_labels)
            if len(location) < 1:
                bundle = Streamlines([])
            else:
                bundle = streamlines[location]

            logging.info("Saving output files ...")
            new_sft = StatefulTractogram(bundle, sft, Space.RASMM)
            save_tractogram(new_sft, out_bundle, bbox_valid_check=False)
            logging.info(out_bundle)


class ClassifyTissueFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "extracttissue"

    def run(
        self,
        input_files,
        bvals_file=None,
        method=None,
        wm_threshold=0.5,
        b0_threshold=50,
        low_signal_threshold=50,
        nclass=None,
        beta=0.1,
        tolerance=1e-05,
        max_iter=100,
        out_dir="",
        out_tissue="tissue_classified.nii.gz",
        out_pve="tissue_classified_pve.nii.gz",
    ):
        """Extract tissue from a volume.

        Parameters
        ----------
        input_files : string
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        bvals_file : string, optional
            Path to the b-values file. Required for 'dam' method.
        method : string, optional
            Method to use for tissue extraction. Options are:
                - 'hmrf': Markov Random Fields modeling approach.
                - 'dam': Directional Average Maps, proposed by :footcite:p:`Cheng2020`.

            'hmrf' method is recommended for T1w images, while 'dam' method is
            recommended for DWI Multishell images (single shell are not recommended).
        wm_threshold : float, optional
            The threshold below which a voxel is considered white matter. For data like
            HCP, threshold of 0.5 proves to be a good choice. For data like cfin, higher
            threshold values like 0.7 or 0.8 are more suitable. Used for 'dam' method.
        b0_threshold : float, optional
            The intensity threshold for a b=0 image. used only for 'dam' method.
        low_signal_threshold : float, optional
            The threshold below which a voxel is considered to have low signal.
            Used only for 'dam' method.
        nclass : int, optional
            Number of desired classes. Used only for 'hmrf' method.
        beta : float, optional
            Smoothing parameter, the higher this number the smoother the
            output will be. Used only for 'hmrf' method.
        tolerance : float, optional
            Value that defines the percentage of change tolerated to
            prevent the ICM loop to stop. Default is 1e-05.
            If you want tolerance check to be disabled put 'tolerance = 0'.
            Used only for 'hmrf' method.
        max_iter : int, optional
            Fixed number of desired iterations. Default is 100.
            This parameter defines the maximum number of iterations the
            algorithm will perform. The loop may terminate early if the
            change in energy sum between iterations falls below the
            threshold defined by `tolerance`. However, if `tolerance` is
            explicitly set to 0, this early stopping mechanism is disabled,
            and the algorithm will run for the specified number of
            iterations unless another stopping criterion is met.
            Used only for 'hmrf' method.
        out_dir : string, optional
            Output directory.
        out_tissue : string, optional
            Name of the tissue volume to be saved.
        out_pve : string, optional
            Name of the pve volume to be saved.

        REFERENCES
        ----------
        .. footbibliography::

        """
        io_it = self.get_io_iterator()

        if not method or method.lower() not in ["hmrf", "dam"]:
            logging.error(
                f"Unknown method '{method}' for tissue extraction. "
                "Choose '--method hmrf' (for T1w) or '--method dam' (for DWI)"
            )
            sys.exit(1)

        prefix = "t1" if method.lower() == "hmrf" else "dwi"
        for i, name in enumerate(self.flat_outputs):
            if name.endswith("tissue_classified.nii.gz"):
                self.flat_outputs[i] = name.replace(
                    "tissue_classified.nii.gz", f"{prefix}_tissue_classified.nii.gz"
                )
            if name.endswith("tissue_classified_pve.nii.gz"):
                self.flat_outputs[i] = name.replace(
                    "tissue_classified_pve.nii.gz",
                    f"{prefix}_tissue_classified_pve.nii.gz",
                )

        self.update_flat_outputs(self.flat_outputs, io_it)

        for fpath, tissue_out_path, opve in io_it:
            logging.info(f"Extracting tissue from {fpath}")

            data, affine = load_nifti(fpath)

            if method.lower() == "hmrf":
                if nclass is None:
                    logging.error(
                        "Number of classes is required for 'hmrf' method. "
                        "For example, Use '--nclass 4' to specify the number of "
                        "classes."
                    )
                    sys.exit(1)
                classifier = TissueClassifierHMRF()
                _, segmentation_final, PVE = classifier.classify(
                    data, nclass, beta, tolerance=tolerance, max_iter=max_iter
                )

                save_nifti(tissue_out_path, segmentation_final, affine)
                save_nifti(opve, PVE, affine)

            elif method.lower() == "dam":
                if bvals_file is None or not os.path.isfile(bvals_file):
                    logging.error("'--bvals filename' is required for 'dam' method")
                    sys.exit(1)

                bvals, _ = read_bvals_bvecs(bvals_file, None)
                wm_mask, gm_mask = dam_classifier(
                    data,
                    bvals,
                    wm_threshold=wm_threshold,
                    b0_threshold=b0_threshold,
                    low_signal_threshold=low_signal_threshold,
                )
                result = np.zeros(wm_mask.shape)
                result[wm_mask] = 1
                result[gm_mask] = 2
                save_nifti(tissue_out_path, result, affine)
                save_nifti(
                    opve, np.stack([wm_mask, gm_mask], axis=-1).astype(np.int32), affine
                )

            logging.info(f"Tissue saved as {tissue_out_path} and PVE as {opve}")

        return io_it



from functools import reduce

import numpy as np
import numpy.testing as npt

from dipy.core.sphere import unit_icosahedron
from dipy.reconst.multi_voxel import CallableArray, _squash, multi_voxel_fit
from dipy.testing.decorators import set_random_number_generator, warning_for_keywords
from dipy.utils.optpkg import optional_package

joblib, has_joblib, _ = optional_package("joblib")
dask, has_dask, _ = optional_package("dask")
ray, has_ray, _ = optional_package("ray")


def test_squash():
    A = np.ones((3, 3), dtype=float)
    B = np.asarray(A, object)
    npt.assert_array_equal(A, _squash(B))
    npt.assert_equal(_squash(B).dtype, A.dtype)

    B[2, 2] = None
    A[2, 2] = 0
    npt.assert_array_equal(A, _squash(B))
    npt.assert_equal(_squash(B).dtype, A.dtype)

    for ijk in np.ndindex(*B.shape):
        B[ijk] = np.ones((2,))
    A = np.ones((3, 3, 2))
    npt.assert_array_equal(A, _squash(B))
    npt.assert_equal(_squash(B).dtype, A.dtype)

    B[2, 2] = None
    A[2, 2] = 0
    npt.assert_array_equal(A, _squash(B))
    npt.assert_equal(_squash(B).dtype, A.dtype)

    # sub-arrays have different shapes ( (3,) and (2,) )
    B[0, 0] = np.ones((3,))
    npt.assert_(_squash(B) is B)

    # Check dtypes for arrays and scalars
    arr_arr = np.zeros((2,), dtype=object)
    scalar_arr = np.zeros((2,), dtype=object)
    numeric_types = [
        getattr(np, dtype)
        for dtype in (
            "int8",
            "byte",
            "int16",
            "short",
            "int32",
            "intc",
            "int_",
            "int64",
            "longlong",
            "uint8",
            "ubyte",
            "uint16",
            "ushort",
            "uint32",
            "uintc",
            "uint",
            "uint64",
            "ulonglong",
            "float16",
            "half",
            "float32",
            "single",
            "float64",
            "double",
            "float96",
            "float128",
            "longdouble",
            "complex64",
            "csingle",
            "complex128",
            "cdouble",
            "complex192",
            "complex256",
            "clongdouble",
        )
        if hasattr(np, dtype)
    ] + [bool]
    for dt0 in numeric_types:
        arr_arr[0] = np.zeros((3,), dtype=dt0)
        scalar_arr[0] = dt0(0)
        for dt1 in numeric_types:
            arr_arr[1] = np.zeros((3,), dtype=dt1)
            npt.assert_equal(_squash(arr_arr).dtype, reduce(np.add, arr_arr).dtype)
            scalar_arr[1] = dt0(1)
            npt.assert_equal(
                _squash(scalar_arr).dtype, reduce(np.add, scalar_arr).dtype
            )

    # Check masks and Nones
    arr = np.ones((3, 4), dtype=float)
    obj_arr = arr.astype(object)
    arr[1, 1] = 99
    obj_arr[1, 1] = None
    npt.assert_array_equal(_squash(obj_arr, mask=None, fill=99), arr)
    msk = arr == 1
    npt.assert_array_equal(_squash(obj_arr, mask=msk, fill=99), arr)
    msk[1, 1] = 1  # unmask None - object array back
    npt.assert_array_equal(_squash(obj_arr, mask=msk, fill=99), obj_arr)
    msk[1, 1] = 0  # remask, back to fill again
    npt.assert_array_equal(_squash(obj_arr, mask=msk, fill=99), arr)
    obj_arr[2, 3] = None  # add another unmasked None, object again
    npt.assert_array_equal(_squash(obj_arr, mask=msk, fill=99), obj_arr)

    # Check array of arrays
    obj_arrs = np.zeros((3,), dtype=object)
    for i in range(3):
        obj_arrs[i] = np.ones((4, 5))
    arr_arrs = np.ones((3, 4, 5))
    # No Nones
    npt.assert_array_equal(_squash(obj_arrs, mask=None, fill=99), arr_arrs)
    # None, implicit masking
    obj_masked = obj_arrs.copy()
    obj_masked[1] = None
    arr_masked = arr_arrs.copy()
    arr_masked[1] = 99
    npt.assert_array_equal(_squash(obj_masked, mask=None, fill=99), arr_masked)
    msk = np.array([1, 0, 1], dtype=bool)  # explicit mask
    npt.assert_array_equal(_squash(obj_masked, mask=msk, fill=99), arr_masked)


def test_CallableArray():
    callarray = CallableArray((2, 3), dtype=object)

    # Test without Nones
    callarray[:] = np.arange
    expected = np.empty([2, 3, 4])
    expected[:] = range(4)
    npt.assert_array_equal(callarray(4), expected)

    # Test with Nones
    callarray[0, 0] = None
    expected[0, 0] = 0
    npt.assert_array_equal(callarray(4), expected)


@set_random_number_generator()
def test_multi_voxel_fit(rng):
    class SillyModel:
        @warning_for_keywords()
        @multi_voxel_fit
        def fit(
            self, data, *, mask=None, another_kwarg=None, kwarg_untouched=True, **kwargs
        ):
            # We want to make sure that all kwargs are passed through to the
            # the fitting procedure
            assert another_kwarg is not None
            # Make sure that an argument that is not passed is still
            # usable in the fitting procedure:
            assert kwarg_untouched
            return SillyFit(model, data)

        def predict(self, S0):
            return np.ones(10) * S0

    class SillyFit:
        def __init__(self, model, data):
            self.model = model
            self.data = data

        model_attr = 2.0

        def odf(self, sphere):
            return np.ones(len(sphere.phi))

        @property
        def directions(self):
            n = rng.integers(0, 10)
            return np.zeros((n, 3))

        def predict(self, S0):
            return np.ones(self.data.shape) * S0

    # Test the single voxel case
    model = SillyModel()
    single_voxel = np.zeros(64)
    fit = model.fit(single_voxel, another_kwarg="foo")
    npt.assert_equal(type(fit), SillyFit)

    # Test without a mask
    many_voxels = np.zeros((2, 3, 4, 64))
    for verbose in [True, False]:
        fit = model.fit(many_voxels, verbose=verbose, another_kwarg="foo")
        expected = np.empty((2, 3, 4))
        expected[:] = 2.0
        npt.assert_array_equal(fit.model_attr, expected)
        expected = np.ones((2, 3, 4, 12))
        npt.assert_array_equal(fit.odf(unit_icosahedron), expected)
        npt.assert_equal(fit.directions.shape, (2, 3, 4))
        S0 = 100.0
        npt.assert_equal(fit.predict(S0=S0), np.ones(many_voxels.shape) * S0)

    # Test with parallelization (using the "serial" dummy engine)
    fit = model.fit(many_voxels, another_kwarg="foo", engine="serial")

    for verbose in [True, False]:
        # Test with single value kwarg, or sequence of kwarg values
        for another_kwarg in ["foo", len(many_voxels) * ["foo"]]:
            # If parallelization engines are installed use them to test:
            if has_joblib:
                fit = model.fit(
                    many_voxels,
                    verbose=verbose,
                    another_kwarg=another_kwarg,
                    engine="joblib",
                )
                npt.assert_equal(fit.predict(S0=S0), np.ones(many_voxels.shape) * S0)

            if has_dask:
                fit = model.fit(
                    many_voxels,
                    verbose=verbose,
                    another_kwarg=another_kwarg,
                    engine="dask",
                )
                npt.assert_equal(fit.predict(S0=S0), np.ones(many_voxels.shape) * S0)

            if has_ray:
                fit = model.fit(
                    many_voxels,
                    verbose=verbose,
                    another_kwarg=another_kwarg,
                    engine="ray",
                )
                npt.assert_equal(fit.predict(S0=S0), np.ones(many_voxels.shape) * S0)

    # Test with a mask
    mask = np.zeros((3, 3, 3)).astype("bool")
    mask[0, 0] = 1
    mask[1, 1] = 1
    mask[2, 2] = 1
    data = np.zeros((3, 3, 3, 64))
    fit = model.fit(data, mask=mask, another_kwarg="foo")
    expected = np.zeros((3, 3, 3))
    expected[0, 0] = 2
    expected[1, 1] = 2
    expected[2, 2] = 2
    npt.assert_array_equal(fit.model_attr, expected)
    odf = fit.odf(unit_icosahedron)
    npt.assert_equal(odf.shape, (3, 3, 3, 12))
    npt.assert_array_equal(odf[~mask], 0)
    npt.assert_array_equal(odf[mask], 1)
    predicted = np.zeros(data.shape)
    predicted[mask] = S0
    npt.assert_equal(fit.predict(S0=S0), predicted)

    # Test fit.shape
    npt.assert_equal(fit.shape, (3, 3, 3))

    # Test indexing into a fit
    npt.assert_equal(type(fit[0, 0, 0]), SillyFit)
    npt.assert_equal(fit[:2, :2, :2].shape, (2, 2, 2))



import numpy as np
from numpy.testing import assert_, assert_equal, assert_raises
import pytest

from dipy.segment.tissue import compute_directional_average, dam_classifier
from dipy.utils.optpkg import optional_package

sklearn, has_sklearn, _ = optional_package("sklearn")
needs_sklearn = pytest.mark.skipif(not has_sklearn, reason="Requires sklearn")


@needs_sklearn
def test_compute_directional_average_valid():
    data = np.array([100, 80, 60, 50, 40, 20, 10])
    bvals = np.array([0, 100, 500, 1000, 1500, 2000, 3000])
    P, V = compute_directional_average(data, bvals)

    assert_(isinstance(P, float), "P should be a float")
    assert_(isinstance(V, float), "V should be a float")
    assert_(P != 0, "P should not be zero")
    assert_(V != 0, "V should not be zero")


@needs_sklearn
def test_compute_directional_average_low_signal():
    data = np.array([20, 10, 5, 3, 2, 1, 1])  # Very low signal
    bvals = np.array([0, 100, 500, 1000, 1500, 2000, 3000])

    P, V = compute_directional_average(data, bvals, low_signal_threshold=50)

    assert_equal(P, 0, "P should be 0 when low signal")
    assert_equal(V, 0, "V should be 0 when low signal")


@needs_sklearn
def test_compute_directional_average_div_by_zero():
    data = np.array([100, 100, 100, 100, 100, 100, 100])
    bvals = np.array([0, 100, 500, 1000, 1500, 2000, 3000])

    P, V = compute_directional_average(data, bvals)

    assert_(np.allclose(P, 0))


@needs_sklearn
def test_dam_classifier_valid():
    data = np.random.rand(3, 3, 3, 7) * 100  # Simulated random data
    bvals = np.array([0, 100, 500, 1000, 1500, 2000, 3000])

    assert_equal(
        data.shape[-1],
        bvals.shape[0],
        "The number of bvals must match the last dimension of data",
    )

    wm_mask, gm_mask = dam_classifier(data, bvals, wm_threshold=0.5)

    assert_equal(wm_mask.shape, (3, 3, 3), "Shape of wm_mask should be (3, 3, 3)")
    assert_equal(gm_mask.shape, (3, 3, 3), "Shape of gm_mask should be (3, 3, 3)")

    data = np.array([100, 80, 60, 50])
    bvals = np.array([0, 0, 100, 100])

    assert_raises(ValueError, dam_classifier, data, bvals, wm_threshold=0.5)



"""Test spherical harmonic models and the tools associated with those models."""

import warnings

import numpy as np
import numpy.linalg as npl
import numpy.testing as npt
from numpy.testing import (
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
    assert_raises,
)

from dipy.core.gradients import gradient_table
from dipy.core.interpolation import NearestNeighborInterpolator
from dipy.core.sphere import Sphere, hemi_icosahedron
from dipy.data import mrtrix_spherical_functions
from dipy.direction.peaks import peak_directions
from dipy.reconst import odf
from dipy.reconst.shm import (
    CsaOdfModel,
    OpdtModel,
    QballModel,
    ResidualBootstrapWrapper,
    SphHarmFit,
    anisotropic_power,
    bootstrap_data_array,
    bootstrap_data_voxel,
    calculate_max_order,
    convert_sh_descoteaux_tournier,
    convert_sh_from_legacy,
    convert_sh_to_full_basis,
    convert_sh_to_legacy,
    descoteaux07_legacy_msg,
    gen_dirac,
    hat,
    lcr_matrix,
    normalize_data,
    order_from_ncoef,
    real_sh_descoteaux,
    real_sh_descoteaux_from_index,
    real_sh_tournier,
    real_sym_sh_basis,
    real_sym_sh_mrtrix,
    sf_to_sh,
    sh_to_sf,
    sh_to_sf_matrix,
    smooth_pinv,
    sph_harm_ind_list,
    spherical_harmonics,
    tournier07_legacy_msg,
)
from dipy.sims.voxel import multi_tensor_odf, single_tensor
from dipy.testing import assert_true


def test_order_from_ncoeff():
    # Just try some out:
    for sh_order_max in [2, 4, 6, 8, 12, 24]:
        m_values, l_values = sph_harm_ind_list(sh_order_max)
        n_coef = m_values.shape[0]
        assert_equal(order_from_ncoef(n_coef), sh_order_max)

        # Try out full basis
        m_full, l_full = sph_harm_ind_list(sh_order_max, full_basis=True)
        n_coef_full = m_full.shape[0]
        assert_equal(order_from_ncoef(n_coef_full, full_basis=True), sh_order_max)


def test_sph_harm_ind_list():
    m_list, l_list = sph_harm_ind_list(8)
    assert_equal(m_list.shape, l_list.shape)
    assert_equal(m_list.shape, (45,))
    assert_true(np.all(np.abs(m_list) <= l_list))
    assert_array_equal(l_list % 2, 0)
    assert_raises(ValueError, sph_harm_ind_list, 1)

    # Test for a full basis
    m_list, l_list = sph_harm_ind_list(8, full_basis=True)
    assert_equal(m_list.shape, l_list.shape)
    # There are (sh_order + 1) * (sh_order + 1) coefficients
    assert_equal(m_list.shape, (81,))
    assert_true(np.all(np.abs(m_list) <= l_list))


def test_real_sh_descoteaux_from_index():
    # Tests derived from tables in
    # https://en.wikipedia.org/wiki/Table_of_spherical_harmonics
    # where real spherical harmonic $Y_l^m$ is defined to be:
    #    Real($Y_l^m$) * sqrt(2) if m > 0
    #    $Y_l^m$                 if m == 0
    #    Imag($Y_l^m$) * sqrt(2) if m < 0

    rsh = real_sh_descoteaux_from_index
    pi = np.pi
    sqrt = np.sqrt
    sin = np.sin
    cos = np.cos

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        assert_array_almost_equal(rsh(0, 0, 0, 0), 0.5 / sqrt(pi))
        assert_array_almost_equal(
            rsh(-2, 2, pi / 5, pi / 3),
            0.25
            * sqrt(15.0 / (2.0 * pi))
            * (sin(pi / 5.0)) ** 2.0
            * cos(0 + 2.0 * pi / 3)
            * sqrt(2),
        )
        assert_array_almost_equal(
            rsh(2, 2, pi / 5, pi / 3),
            -1
            * 0.25
            * sqrt(15.0 / (2.0 * pi))
            * (sin(pi / 5.0)) ** 2.0
            * sin(0 - 2.0 * pi / 3)
            * sqrt(2),
        )
        assert_array_almost_equal(
            rsh(-2, 2, pi / 2, pi),
            0.25
            * sqrt(15 / (2.0 * pi))
            * cos(2.0 * pi)
            * sin(pi / 2.0) ** 2.0
            * sqrt(2),
        )
        assert_array_almost_equal(
            rsh(2, 4, pi / 3.0, pi / 4.0),
            -1
            * (3.0 / 8.0)
            * sqrt(5.0 / (2.0 * pi))
            * sin(0 - 2.0 * pi / 4.0)
            * sin(pi / 3.0) ** 2.0
            * (7.0 * cos(pi / 3.0) ** 2.0 - 1)
            * sqrt(2),
        )
        assert_array_almost_equal(
            rsh(-4, 4, pi / 6.0, pi / 8.0),
            (3.0 / 16.0)
            * sqrt(35.0 / (2.0 * pi))
            * cos(0 + 4.0 * pi / 8.0)
            * sin(pi / 6.0) ** 4.0
            * sqrt(2),
        )
        assert_array_almost_equal(
            rsh(4, 4, pi / 6.0, pi / 8.0),
            -1
            * (3.0 / 16.0)
            * sqrt(35.0 / (2.0 * pi))
            * sin(0 - 4.0 * pi / 8.0)
            * sin(pi / 6.0) ** 4.0
            * sqrt(2),
        )

    aa = np.ones((3, 1, 1, 1))
    bb = np.ones((1, 4, 1, 1))
    cc = np.ones((1, 1, 5, 1))
    dd = np.ones((1, 1, 1, 6))

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        assert_equal(rsh(aa, bb, cc, dd).shape, (3, 4, 5, 6))


def test_gen_dirac():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        sh = gen_dirac(np.array([0]), np.array([0]), np.array([0]), np.array([0]))

    assert_true(np.abs(sh[0] - 1.0 / np.sqrt(4.0 * np.pi)) < 0.0001)


def test_real_sym_sh_mrtrix():
    coef, expected, sphere = mrtrix_spherical_functions()

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        basis, m_values, l_values = real_sym_sh_mrtrix(8, sphere.theta, sphere.phi)

    npt.assert_equal(len(w), 2)
    npt.assert_(issubclass(w[0].category, DeprecationWarning))
    npt.assert_(
        "dipy.reconst.shm.real_sym_sh_mrtrix is deprecated, Please use "
        "dipy.reconst.shm.real_sh_tournier instead" in str(w[0].message)
    )
    npt.assert_(issubclass(w[1].category, PendingDeprecationWarning))
    npt.assert_(tournier07_legacy_msg in str(w[1].message))

    func = np.dot(coef, basis.T)
    assert_array_almost_equal(func, expected, 4)


def test_real_sym_sh_basis():
    new_order = [0, 5, 4, 3, 2, 1, 14, 13, 12, 11, 10, 9, 8, 7, 6]
    sphere = hemi_icosahedron.subdivide(n=2)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")

        basis, m_values, l_values = real_sym_sh_mrtrix(4, sphere.theta, sphere.phi)

    expected = basis[:, new_order]
    expected *= np.where(m_values == 0, 1.0, np.sqrt(2))

    with warnings.catch_warnings(record=True) as w:
        descoteaux07_basis, m_values, l_values = real_sym_sh_basis(
            4, sphere.theta, sphere.phi
        )

    npt.assert_equal(len(w), 2)
    npt.assert_(issubclass(w[0].category, DeprecationWarning))
    npt.assert_(
        "dipy.reconst.shm.real_sym_sh_basis is deprecated, Please use "
        "dipy.reconst.shm.real_sh_descoteaux instead" in str(w[0].message)
    )
    npt.assert_(issubclass(w[1].category, PendingDeprecationWarning))
    npt.assert_(descoteaux07_legacy_msg in str(w[1].message))

    assert_array_almost_equal(descoteaux07_basis, expected)


def test_real_sh_descoteaux1():
    # This test should do for now
    # The tournier07 basis should be the same as re-ordering and re-scaling the
    # descoteaux07 basis
    new_order = [0, 5, 4, 3, 2, 1, 14, 13, 12, 11, 10, 9, 8, 7, 6]
    sphere = hemi_icosahedron.subdivide(n=2)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        basis, m_values, l_values = real_sym_sh_mrtrix(4, sphere.theta, sphere.phi)

    expected = basis[:, new_order]
    expected *= np.where(m_values == 0, 1.0, np.sqrt(2))

    with warnings.catch_warnings(record=True) as w:
        descoteaux07_basis, m_values, l_values = real_sh_descoteaux(
            4, sphere.theta, sphere.phi
        )

    npt.assert_equal(len(w), 1)
    npt.assert_(issubclass(w[0].category, PendingDeprecationWarning))
    npt.assert_(descoteaux07_legacy_msg in str(w[0].message))

    assert_array_almost_equal(descoteaux07_basis, expected)


def test_real_sh_tournier():
    vertices = hemi_icosahedron.subdivide(n=2).vertices
    mevals = np.array([[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]])
    angles = [(0, 0), (60, 0)]
    odf = multi_tensor_odf(vertices, mevals, angles, [50, 50])

    mevals = np.array([[0.0015, 0.0003, 0.0003]])
    angles = [(0, 0)]
    odf2 = multi_tensor_odf(-vertices, mevals, angles, [100])

    sphere = Sphere(xyz=np.vstack((vertices, -vertices)))
    # Asymmetric spherical function with 162 coefficients
    sf = np.append(odf, odf2)

    # In order for our approximation to be precise enough, we
    # will use a SH basis of orders up to 10 (121 coefficients)

    with warnings.catch_warnings(record=True) as w:
        B, m_values, l_values = real_sh_tournier(
            10, sphere.theta, sphere.phi, full_basis=True
        )

    npt.assert_equal(len(w), 1)
    npt.assert_(issubclass(w[0].category, PendingDeprecationWarning))
    npt.assert_(tournier07_legacy_msg in str(w[0].message))

    invB = smooth_pinv(B, L=np.zeros_like(l_values))
    sh_coefs = np.dot(invB, sf)
    sf_approx = np.dot(B, sh_coefs)

    assert_array_almost_equal(sf_approx, sf, 2)


def test_real_sh_descoteaux2():
    vertices = hemi_icosahedron.subdivide(n=2).vertices
    mevals = np.array([[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]])
    angles = [(0, 0), (60, 0)]
    odf = multi_tensor_odf(vertices, mevals, angles, [50, 50])

    mevals = np.array([[0.0015, 0.0003, 0.0003]])
    angles = [(0, 0)]
    odf2 = multi_tensor_odf(-vertices, mevals, angles, [100])

    sphere = Sphere(xyz=np.vstack((vertices, -vertices)))
    # Asymmetric spherical function with 162 coefficients
    sf = np.append(odf, odf2)

    # In order for our approximation to be precise enough, we
    # will use a SH basis of orders up to 10 (121 coefficients)

    with warnings.catch_warnings(record=True) as w:
        B, m_values, l_values = real_sh_descoteaux(
            10, sphere.theta, sphere.phi, full_basis=True
        )

    npt.assert_equal(len(w), 1)
    npt.assert_(issubclass(w[0].category, PendingDeprecationWarning))
    npt.assert_(descoteaux07_legacy_msg in str(w[0].message))

    invB = smooth_pinv(B, L=np.zeros_like(l_values))
    sh_coefs = np.dot(invB, sf)
    sf_approx = np.dot(B, sh_coefs)

    assert_array_almost_equal(sf_approx, sf, 2)


def test_sh_to_sf_matrix():
    sphere = Sphere(xyz=hemi_icosahedron.vertices)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        B1, invB1 = sh_to_sf_matrix(sphere)

        B2, m_values, l_values = real_sh_descoteaux(4, sphere.theta, sphere.phi)

    invB2 = smooth_pinv(B2, L=np.zeros_like(l_values))

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        B3 = sh_to_sf_matrix(sphere, return_inv=False)

    assert_array_almost_equal(B1, B2.T)
    assert_array_almost_equal(invB1, invB2.T)
    assert_array_almost_equal(B3, B1)
    assert_raises(ValueError, sh_to_sf_matrix, sphere, basis_type="")


def test_smooth_pinv():
    hemi = hemi_icosahedron.subdivide(n=2)
    m_values, l_values = sph_harm_ind_list(4)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        B = real_sh_descoteaux_from_index(
            m_values, l_values, hemi.theta[:, None], hemi.phi[:, None]
        )

    L = np.zeros(len(m_values))
    C = smooth_pinv(B, L)
    D = np.dot(npl.inv(np.dot(B.T, B)), B.T)
    assert_array_almost_equal(C, D)

    L = l_values * (l_values + 1) * 0.05
    C = smooth_pinv(B, L)
    L = np.diag(L)
    D = np.dot(npl.inv(np.dot(B.T, B) + L * L), B.T)

    assert_array_almost_equal(C, D)

    L = np.arange(len(l_values)) * 0.05
    C = smooth_pinv(B, L)
    L = np.diag(L)
    D = np.dot(npl.inv(np.dot(B.T, B) + L * L), B.T)
    assert_array_almost_equal(C, D)


def test_normalize_data():
    sig = np.arange(1, 66)[::-1]

    where_b0 = np.zeros(65, "bool")
    where_b0[0] = True
    assert_raises(ValueError, normalize_data, sig, where_b0, out=sig)

    norm_sig = normalize_data(sig, where_b0, min_signal=1)
    assert_array_almost_equal(norm_sig, sig / 65.0)
    norm_sig = normalize_data(sig, where_b0, min_signal=5)
    assert_array_almost_equal(norm_sig[-5:], 5 / 65.0)

    where_b0[[0, 1]] = [True, True]
    norm_sig = normalize_data(sig, where_b0, min_signal=1)
    assert_array_almost_equal(norm_sig, sig / 64.5)
    norm_sig = normalize_data(sig, where_b0, min_signal=5)
    assert_array_almost_equal(norm_sig[-5:], 5 / 64.5)

    sig = sig * np.ones((2, 3, 1))

    where_b0[[0, 1]] = [True, False]
    norm_sig = normalize_data(sig, where_b0, min_signal=1)
    assert_array_almost_equal(norm_sig, sig / 65.0)
    norm_sig = normalize_data(sig, where_b0, min_signal=5)
    assert_array_almost_equal(norm_sig[..., -5:], 5 / 65.0)

    where_b0[[0, 1]] = [True, True]
    norm_sig = normalize_data(sig, where_b0, min_signal=1)
    assert_array_almost_equal(norm_sig, sig / 64.5)
    norm_sig = normalize_data(sig, where_b0, min_signal=5)
    assert_array_almost_equal(norm_sig[..., -5:], 5 / 64.5)


def make_fake_signal():
    hemisphere = hemi_icosahedron.subdivide(n=2)
    bvecs = np.concatenate(([[0, 0, 0]], hemisphere.vertices))
    bvals = np.zeros(len(bvecs)) + 2000
    bvals[0] = 0
    gtab = gradient_table(bvals, bvecs=bvecs)

    evals = np.array([[2.1, 0.2, 0.2], [0.2, 2.1, 0.2]]) * 10**-3
    evecs0 = np.eye(3)
    evecs1 = evecs0
    a = evecs0[0]
    b = evecs1[1]
    S1 = single_tensor(gtab, 0.55, evals=evals[0], evecs=evecs0)
    S2 = single_tensor(gtab, 0.45, evals=evals[1], evecs=evecs1)
    return S1 + S2, gtab, np.vstack([a, b])


class TestQballModel:
    model = QballModel

    def test_single_voxel_fit(self):
        signal, gtab, expected = make_fake_signal()
        sphere = hemi_icosahedron.subdivide(n=4)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            model = self.model(
                gtab, sh_order_max=4, min_signal=1e-5, assume_normed=True
            )

        fit = model.fit(signal)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            odf = fit.odf(sphere)

        assert_equal(odf.shape, sphere.phi.shape)
        directions, _, _ = peak_directions(odf, sphere)
        # Check the same number of directions
        n = len(expected)
        assert_equal(len(directions), n)
        # Check directions are unit vectors
        cos_similarity = (directions * directions).sum(-1)
        assert_array_almost_equal(cos_similarity, np.ones(n))
        # Check the directions == expected or -expected
        cos_similarity = (directions * expected).sum(-1)
        assert_array_almost_equal(abs(cos_similarity), np.ones(n))

        # Test normalize data
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            model = self.model(
                gtab, sh_order_max=4, min_signal=1e-5, assume_normed=False
            )

        fit = model.fit(signal * 5)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            odf_with_norm = fit.odf(sphere)

        assert_array_almost_equal(odf, odf_with_norm)

    def test_mulit_voxel_fit(self):
        signal, gtab, expected = make_fake_signal()
        sphere = hemi_icosahedron
        nd_signal = np.vstack([signal, signal])

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            model = self.model(
                gtab, sh_order_max=4, min_signal=1e-5, assume_normed=True
            )

        fit = model.fit(nd_signal)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            odf = fit.odf(sphere)

        assert_equal(odf.shape, (2,) + sphere.phi.shape)

        # Test fitting with mask, where mask is False odf should be 0
        fit = model.fit(nd_signal, mask=[False, True])

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            odf = fit.odf(sphere)

        assert_array_equal(odf[0], 0.0)

    def test_sh_order(self):
        signal, gtab, expected = make_fake_signal()

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            model = self.model(gtab, sh_order_max=4, min_signal=1e-5)

        assert_equal(model.B.shape[1], 15)
        assert_equal(max(model.l_values), 4)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            model = self.model(gtab, sh_order_max=6, min_signal=1e-5)

        assert_equal(model.B.shape[1], 28)
        assert_equal(max(model.l_values), 6)

    def test_gfa(self):
        signal, gtab, expected = make_fake_signal()
        signal = np.ones((2, 3, 4, 1)) * signal
        sphere = hemi_icosahedron.subdivide(n=3)

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            model = self.model(gtab, 6, min_signal=1e-5)

        fit = model.fit(signal)
        gfa_shm = fit.gfa

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            gfa_odf = odf.gfa(fit.odf(sphere))

        assert_array_almost_equal(gfa_shm, gfa_odf, 3)

        # gfa should be 0 if all coefficients are 0 (masked areas)
        mask = np.zeros(signal.shape[:-1])
        fit = model.fit(signal, mask=mask)
        assert_array_equal(fit.gfa, 0)

    def test_min_signal_default(self):
        signal, gtab, expected = make_fake_signal()

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            model_default = self.model(gtab, 4)

        shm_default = model_default.fit(signal).shm_coeff

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=descoteaux07_legacy_msg,
                category=PendingDeprecationWarning,
            )

            model_correct = self.model(gtab, 4, min_signal=1e-5)

        shm_correct = model_correct.fit(signal).shm_coeff
        assert_equal(shm_default, shm_correct)


def test_SphHarmFit():
    coef = np.zeros((3, 4, 5, 45))
    mask = np.zeros((3, 4, 5), dtype=bool)

    fit = SphHarmFit(None, coef, mask)
    item = fit[0, 0, 0]
    assert_equal(item.shape, ())
    data = fit[0]
    assert_equal(data.shape, (4, 5))
    data = fit[:, :, 0]
    assert_equal(data.shape, (3, 4))


class TestOpdtModel(TestQballModel):
    model = OpdtModel


class TestCsaOdfModel(TestQballModel):
    model = CsaOdfModel


def test_hat_and_lcr():
    hemi = hemi_icosahedron.subdivide(n=3)
    m_values, l_values = sph_harm_ind_list(8)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        B = real_sh_descoteaux_from_index(
            m_values, l_values, hemi.theta[:, None], hemi.phi[:, None]
        )

    H = hat(B)
    B_hat = np.dot(H, B)
    assert_array_almost_equal(B, B_hat)

    R = lcr_matrix(H)
    d = np.arange(len(hemi.theta))
    r = d - np.dot(H, d)
    lev = np.sqrt(1 - H.diagonal())
    r /= lev
    r -= r.mean()

    r2 = np.dot(R, d)
    assert_array_almost_equal(r, r2)

    r3 = np.dot(d, R.T)
    assert_array_almost_equal(r, r3)


def test_bootstrap_array():
    B = np.array([[4, 5, 7, 4, 2.0], [4, 6, 2, 3, 6.0]])
    H = hat(B.T)

    R = np.zeros((5, 5))
    d = np.arange(1, 6)
    dhat = np.dot(H, d)

    assert_array_almost_equal(bootstrap_data_voxel(dhat, H, R), dhat)
    assert_array_almost_equal(bootstrap_data_array(dhat, H, R), dhat)


def test_ResidualBootstrapWrapper():
    B = np.array([[4, 5, 7, 4, 2.0], [4, 6, 2, 3, 6.0]])
    B = B.T
    H = hat(B)
    d = np.arange(10) / 8.0
    d.shape = (2, 5)
    dhat = np.dot(d, H)
    signal_object = NearestNeighborInterpolator(dhat, (1,))
    ms = 0.2
    where_dwi = np.ones(len(H), dtype=bool)

    boot_obj = ResidualBootstrapWrapper(signal_object, B, where_dwi, min_signal=ms)
    assert_array_almost_equal(boot_obj[0], dhat[0].clip(ms, 1))
    assert_array_almost_equal(boot_obj[1], dhat[1].clip(ms, 1))

    dhat = np.column_stack([[0.6, 0.7], dhat])
    signal_object = NearestNeighborInterpolator(dhat, (1,))
    where_dwi = np.concatenate([[False], where_dwi])
    boot_obj = ResidualBootstrapWrapper(signal_object, B, where_dwi, min_signal=ms)
    assert_array_almost_equal(boot_obj[0], dhat[0].clip(ms, 1))
    assert_array_almost_equal(boot_obj[1], dhat[1].clip(ms, 1))


def test_sf_to_sh():
    # Subdividing a hemi_icosahedron twice produces 81 unique points, which
    # is more than enough to fit a order 8 (45 coefficients) spherical harmonic
    hemisphere = hemi_icosahedron.subdivide(n=2)
    mevals = np.array([[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]])
    angles = [(0, 0), (60, 0)]
    odf = multi_tensor_odf(hemisphere.vertices, mevals, angles, [50, 50])

    # 1D case with the 2 symmetric bases functions
    # Tournier basis
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )

        odf_sh = sf_to_sh(odf, hemisphere, sh_order_max=8, basis_type="tournier07")
        odf_reconst = sh_to_sf(
            odf_sh, hemisphere, sh_order_max=8, basis_type="tournier07"
        )

    assert_array_almost_equal(odf, odf_reconst, 2)

    # Legacy definition
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )

        odf_sh = sf_to_sh(
            odf, hemisphere, sh_order_max=8, basis_type="tournier07", legacy=True
        )
        odf_reconst = sh_to_sf(
            odf_sh, hemisphere, sh_order_max=8, basis_type="tournier07", legacy=True
        )

    assert_array_almost_equal(odf, odf_reconst, 2)

    # Descoteaux basis
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        odf_sh = sf_to_sh(odf, hemisphere, sh_order_max=8, basis_type="descoteaux07")
        odf_reconst = sh_to_sf(
            odf_sh, hemisphere, sh_order_max=8, basis_type="descoteaux07"
        )

    assert_array_almost_equal(odf, odf_reconst, 2)

    # Legacy definition
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        odf_sh = sf_to_sh(
            odf, hemisphere, sh_order_max=8, basis_type="descoteaux07", legacy=True
        )
        odf_reconst = sh_to_sf(
            odf_sh, hemisphere, sh_order_max=8, basis_type="descoteaux07", legacy=True
        )

    assert_array_almost_equal(odf, odf_reconst, 2)

    # We now create an asymmetric signal
    # to try out our full SH basis
    mevals = np.array([[0.0015, 0.0003, 0.0003]])
    angles = [(0, 0)]
    odf2 = multi_tensor_odf(hemisphere.vertices, mevals, angles, [100])

    # We simulate our asymmetric signal by using a different ODF
    # per hemisphere. The sphere used is a concatenation of the
    # vertices of our hemisphere, for a total of 162 vertices.
    sphere = Sphere(xyz=np.vstack((hemisphere.vertices, -hemisphere.vertices)))
    asym_odf = np.append(odf, odf2)

    # Try out full bases with order 10 (121 coefficients)
    # Tournier basis
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )

        odf_sh = sf_to_sh(
            asym_odf, sphere, sh_order_max=10, basis_type="tournier07", full_basis=True
        )
        odf_reconst = sh_to_sf(
            odf_sh, sphere, sh_order_max=10, basis_type="tournier07", full_basis=True
        )

    assert_array_almost_equal(odf_reconst, asym_odf, 2)

    # Legacy definition
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )

        odf_sh = sf_to_sh(
            asym_odf,
            sphere,
            sh_order_max=10,
            basis_type="tournier07",
            full_basis=True,
            legacy=True,
        )
        odf_reconst = sh_to_sf(
            odf_sh,
            sphere,
            sh_order_max=10,
            basis_type="tournier07",
            full_basis=True,
            legacy=True,
        )

    assert_array_almost_equal(odf_reconst, asym_odf, 2)

    # Descoteaux basis
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        odf_sh = sf_to_sh(
            asym_odf,
            sphere,
            sh_order_max=10,
            basis_type="descoteaux07",
            full_basis=True,
        )
        odf_reconst = sh_to_sf(
            odf_sh, sphere, sh_order_max=10, basis_type="descoteaux07", full_basis=True
        )

    assert_array_almost_equal(odf_reconst, asym_odf, 2)

    # Legacy definition
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        odf_sh = sf_to_sh(
            asym_odf,
            sphere,
            sh_order_max=10,
            basis_type="descoteaux07",
            full_basis=True,
            legacy=True,
        )
        odf_reconst = sh_to_sf(
            odf_sh,
            sphere,
            sh_order_max=10,
            basis_type="descoteaux07",
            full_basis=True,
            legacy=True,
        )

    assert_array_almost_equal(odf_reconst, asym_odf, 2)

    # An invalid basis name should raise an error
    assert_raises(ValueError, sh_to_sf, odf, hemisphere, basis_type="")
    assert_raises(ValueError, sf_to_sh, odf_sh, hemisphere, basis_type="")

    # 2D case
    odf2d = np.vstack((odf, odf))

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        odf2d_sh = sf_to_sh(odf2d, hemisphere, sh_order_max=8)
        odf2d_sf = sh_to_sf(odf2d_sh, hemisphere, sh_order_max=8)

    assert_array_almost_equal(odf2d, odf2d_sf, 2)


def test_faster_sph_harm():
    sh_order_max = 8
    m_values, l_values = sph_harm_ind_list(sh_order_max)
    theta = np.array(
        [
            1.61491146,
            0.76661665,
            0.11976141,
            1.20198246,
            1.74066314,
            1.5925956,
            2.13022055,
            0.50332859,
            1.19868988,
            0.78440679,
            0.50686938,
            0.51739718,
            1.80342999,
            0.73778957,
            2.28559395,
            1.29569064,
            1.86877091,
            0.39239191,
            0.54043037,
            1.61263047,
            0.72695314,
            1.90527318,
            1.58186125,
            0.23130073,
            2.51695237,
            0.99835604,
            1.2883426,
            0.48114057,
            1.50079318,
            1.07978624,
            1.9798903,
            2.36616966,
            2.49233299,
            2.13116602,
            1.36801518,
            1.32932608,
            0.95926683,
            1.070349,
            0.76355762,
            2.07148422,
            1.50113501,
            1.49823314,
            0.89248164,
            0.22187079,
            1.53805373,
            1.9765295,
            1.13361568,
            1.04908355,
            1.68737368,
            1.91732452,
            1.01937457,
            1.45839,
            0.49641525,
            0.29087155,
            0.52824641,
            1.29875871,
            1.81023541,
            1.17030475,
            2.24953206,
            1.20280498,
            0.76399964,
            2.16109722,
            0.79780421,
            0.87154509,
        ]
    )

    phi = np.array(
        [
            -1.5889514,
            -3.11092733,
            -0.61328674,
            -2.4485381,
            2.88058822,
            2.02165946,
            -1.99783366,
            2.71235211,
            1.41577992,
            -2.29413676,
            -2.24565773,
            -1.55548635,
            2.59318232,
            -1.84672472,
            -2.33710739,
            2.12111948,
            1.87523722,
            -1.05206575,
            -2.85381987,
            -2.22808984,
            2.3202034,
            -2.19004474,
            -1.90358372,
            2.14818373,
            3.1030696,
            -2.86620183,
            -2.19860123,
            -0.45468447,
            -3.0034923,
            1.73345011,
            -2.51716288,
            2.49961525,
            -2.68782986,
            2.69699056,
            1.78566133,
            -1.59119705,
            -2.53378963,
            -2.02476738,
            1.36924987,
            2.17600517,
            2.38117241,
            2.99021511,
            -1.4218007,
            -2.44016802,
            -2.52868164,
            3.01531658,
            2.50093627,
            -1.70745826,
            -2.7863931,
            -2.97359741,
            2.17039906,
            2.68424643,
            1.77896086,
            0.45476215,
            0.99734418,
            -2.73107896,
            2.28815009,
            2.86276506,
            3.09450274,
            -3.09857384,
            -1.06955885,
            -2.83826831,
            1.81932195,
            2.81296654,
        ]
    )

    sh = spherical_harmonics(
        m_values, l_values, theta[:, None], phi[:, None], use_scipy=False
    )
    sh2 = spherical_harmonics(
        m_values, l_values, theta[:, None], phi[:, None], use_scipy=True
    )

    assert_array_almost_equal(sh, sh2, 8)
    sh = spherical_harmonics(
        m_values, l_values, theta[:, None], phi[:, None], use_scipy=False
    )
    assert_array_almost_equal(sh, sh2, 8)


def test_anisotropic_power():
    for n_coeffs in [6, 15, 28, 45, 66, 91]:
        for norm_factor in [0.0005, 0.00001]:
            # Create some really simple cases:
            coeffs = np.ones((3, n_coeffs))
            max_order = calculate_max_order(coeffs.shape[-1])
            # For the case where all coeffs == 1, the ap is simply log of the
            # number of even orders up to the maximal order:
            analytic = np.log(len(range(2, max_order + 2, 2))) - np.log(norm_factor)

            answers = [analytic] * 3
            apvals = anisotropic_power(coeffs, norm_factor=norm_factor)
            assert_array_almost_equal(apvals, answers)
            # Test that this works for single voxel arrays as well:
            assert_array_almost_equal(
                anisotropic_power(coeffs[1], norm_factor=norm_factor), answers[1]
            )

    # Test that even when we look at an all-zeros voxel, this
    # avoids a log-of-zero warning:
    with warnings.catch_warnings(record=True) as w:
        assert_equal(anisotropic_power(np.zeros(6)), 0)
        assert len(w) == 0


def test_calculate_max_order():
    """Based on the table in:
    https://jdtournier.github.io/mrtrix-0.2/tractography/preprocess.html
    """
    orders = [2, 4, 6, 8, 10, 12]
    n_coeffs_sym = [6, 15, 28, 45, 66, 91]
    # n = (R + 1)^2 for a full basis
    n_coeffs_full = [9, 25, 49, 81, 121, 169]
    for o, n_sym, n_full in zip(orders, n_coeffs_sym, n_coeffs_full):
        assert_equal(calculate_max_order(n_sym), o)
        assert_equal(calculate_max_order(n_full, full_basis=True), o)

    assert_raises(ValueError, calculate_max_order, 29)
    assert_raises(ValueError, calculate_max_order, 29, full_basis=True)


def test_convert_sh_to_full_basis():
    hemisphere = hemi_icosahedron.subdivide(n=2)
    mevals = np.array([[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]])
    angles = [(0, 0), (60, 0)]
    odf = multi_tensor_odf(hemisphere.vertices, mevals, angles, [50, 50])

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        sh_coeffs = sf_to_sh(odf, hemisphere, sh_order_max=8)

    full_sh_coeffs = convert_sh_to_full_basis(sh_coeffs)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        odf_reconst = sh_to_sf(
            full_sh_coeffs, hemisphere, sh_order_max=8, full_basis=True
        )

    assert_array_almost_equal(odf, odf_reconst, 2)


def test_convert_sh_from_legacy():
    hemisphere = hemi_icosahedron.subdivide(n=2)
    mevals = np.array([[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]])
    angles = [(0, 0), (60, 0)]
    odf = multi_tensor_odf(hemisphere.vertices, mevals, angles, [50, 50])

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        sh_coeffs = sf_to_sh(odf, hemisphere, sh_order_max=8, legacy=True)

    converted_coeffs = convert_sh_from_legacy(sh_coeffs, "descoteaux07")
    expected_coeffs = sf_to_sh(odf, hemisphere, sh_order_max=8, legacy=False)

    assert_array_almost_equal(converted_coeffs, expected_coeffs, 2)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )

        sh_coeffs = sf_to_sh(
            odf, hemisphere, sh_order_max=8, basis_type="tournier07", legacy=True
        )
    converted_coeffs = convert_sh_from_legacy(sh_coeffs, "tournier07")
    expected_coeffs = sf_to_sh(
        odf, hemisphere, sh_order_max=8, basis_type="tournier07", legacy=False
    )

    assert_array_almost_equal(converted_coeffs, expected_coeffs, 2)

    # 2D case
    odfs = np.array([odf, odf])

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )

        sh_coeffs = sf_to_sh(
            odfs,
            hemisphere,
            sh_order_max=8,
            basis_type="tournier07",
            legacy=True,
            full_basis=True,
        )

    converted_coeffs = convert_sh_from_legacy(sh_coeffs, "tournier07", full_basis=True)
    expected_coeffs = sf_to_sh(
        odfs,
        hemisphere,
        sh_order_max=8,
        basis_type="tournier07",
        full_basis=True,
        legacy=False,
    )

    assert_array_almost_equal(converted_coeffs, expected_coeffs, 2)
    assert_raises(ValueError, convert_sh_from_legacy, sh_coeffs, "", full_basis=True)


def test_convert_sh_to_legacy():
    hemisphere = hemi_icosahedron.subdivide(n=2)
    mevals = np.array([[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]])
    angles = [(0, 0), (60, 0)]
    odf = multi_tensor_odf(hemisphere.vertices, mevals, angles, [50, 50])

    sh_coeffs = sf_to_sh(odf, hemisphere, sh_order_max=8, legacy=False)
    converted_coeffs = convert_sh_to_legacy(sh_coeffs, "descoteaux07")

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )

        expected_coeffs = sf_to_sh(odf, hemisphere, sh_order_max=8, legacy=True)

    assert_array_almost_equal(converted_coeffs, expected_coeffs, 2)

    sh_coeffs = sf_to_sh(
        odf, hemisphere, sh_order_max=8, basis_type="tournier07", legacy=False
    )
    converted_coeffs = convert_sh_to_legacy(sh_coeffs, "tournier07")

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )

        expected_coeffs = sf_to_sh(
            odf, hemisphere, sh_order_max=8, basis_type="tournier07", legacy=True
        )

    assert_array_almost_equal(converted_coeffs, expected_coeffs, 2)

    # 2D case
    odfs = np.array([odf, odf])
    sh_coeffs = sf_to_sh(
        odfs,
        hemisphere,
        sh_order_max=8,
        basis_type="tournier07",
        full_basis=True,
        legacy=False,
    )
    converted_coeffs = convert_sh_to_legacy(sh_coeffs, "tournier07", full_basis=True)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message=tournier07_legacy_msg, category=PendingDeprecationWarning
        )

        expected_coeffs = sf_to_sh(
            odfs,
            hemisphere,
            sh_order_max=8,
            basis_type="tournier07",
            legacy=True,
            full_basis=True,
        )

    assert_array_almost_equal(converted_coeffs, expected_coeffs, 2)
    assert_raises(ValueError, convert_sh_to_legacy, sh_coeffs, "", full_basis=True)


def test_convert_sh_descoteaux_tournier():
    # case: max degree zero
    sh_coeffs = np.array([1.54])  # there is only an l=0,m=0 coefficient
    converted_coeffs = convert_sh_descoteaux_tournier(sh_coeffs)
    assert_array_equal(converted_coeffs, sh_coeffs)

    # case: max degree 4
    sh_coeffs = np.array(
        [
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
        ],
        dtype=float,
    )
    # expected result is that there is a swap m <--> -m
    expected_coeffs = np.array(
        [
            1,
            6,
            5,
            4,
            3,
            2,
            15,
            14,
            13,
            12,
            11,
            10,
            9,
            8,
            7,
        ],
        dtype=float,
    )
    converted_coeffs = convert_sh_descoteaux_tournier(sh_coeffs)
    assert_array_equal(converted_coeffs, expected_coeffs)

    # case: max degree 4 but with more axes
    dim0 = 2
    dim1 = 3
    sh_coeffs_grid = np.array(
        [np.linspace(10 * i, 10 * i + 1, 6) for i in range(6)]
    ).reshape(dim0, dim1, 6)
    converted_coeffs_grid = convert_sh_descoteaux_tournier(sh_coeffs_grid)
    assert_equal(sh_coeffs_grid.shape, converted_coeffs_grid.shape)
    for i0 in range(dim0):
        for i1 in range(dim1):
            shc = sh_coeffs_grid[i0, i1]  # shc is short for "sh_coeffs"
            converted_coeffs = converted_coeffs_grid[i0, i1]
            expected_coeffs = np.array(
                [
                    shc[0],
                    shc[5],
                    shc[4],
                    shc[3],
                    shc[2],
                    shc[1],
                ]
            )
            assert_array_equal(converted_coeffs, expected_coeffs)



import numpy as np

from dipy.segment.mrf import ConstantObservationModel, IteratedConditionalModes
from dipy.sims.voxel import add_noise
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

sklearn, has_sklearn, _ = optional_package("sklearn")
linear_model, _, _ = optional_package("sklearn.linear_model")


class TissueClassifierHMRF:
    """
    This class contains the methods for tissue classification using the
    Markov Random Fields modeling approach.
    """

    @warning_for_keywords()
    def __init__(self, *, save_history=False, verbose=True):
        self.save_history = save_history
        self.segmentations = []
        self.pves = []
        self.energies = []
        self.energies_sum = []
        self.verbose = verbose

    @warning_for_keywords()
    def classify(self, image, nclasses, beta, *, tolerance=1e-05, max_iter=100):
        """
        This method uses the Maximum a posteriori - Markov Random Field
        approach for segmentation by using the Iterative Conditional Modes
        and Expectation Maximization to estimate the parameters.

        Parameters
        ----------
        image : ndarray,
            3D structural image.
        nclasses : int,
            Number of desired classes.
        beta : float,
            Smoothing parameter, the higher this number the smoother the
            output will be.
        tolerance: float, optional
            Value that defines the percentage of change tolerated to
            prevent the ICM loop to stop. Default is 1e-05.
            If you want tolerance check to be disabled put 'tolerance = 0'.
        max_iter : int, optional
            Fixed number of desired iterations. Default is 100.
            This parameter defines the maximum number of iterations the
            algorithm will perform. The loop may terminate early if the
            change in energy sum between iterations falls below the
            threshold defined by `tolerance`. However, if `tolerance` is
            explicitly set to 0, this early stopping mechanism is disabled,
            and the algorithm will run for the specified number of
            iterations unless another stopping criterion is met.

        Returns
        -------
        initial_segmentation : ndarray,
            3D segmented image with all tissue types specified in nclasses.
        final_segmentation : ndarray,
            3D final refined segmentation containing all tissue types.
        PVE : ndarray,
            3D probability map of each tissue type.
        """
        nclasses += 1  # One extra class for the background
        energy_sum = [1e-05]

        com = ConstantObservationModel()
        icm = IteratedConditionalModes()

        if image.max() > 1:
            image = np.interp(image, [0, image.max()], [0.0, 1.0])

        mu, sigmasq = com.initialize_param_uniform(image, nclasses)
        p = np.argsort(mu)
        mu = mu[p]
        sigmasq = sigmasq[p]

        neglogl = com.negloglikelihood(image, mu, sigmasq, nclasses)
        seg_init = icm.initialize_maximum_likelihood(neglogl)

        mu, sigmasq = com.seg_stats(image, seg_init, nclasses)

        zero = np.zeros_like(image) + 0.001
        zero_noise = add_noise(zero, 10000, 1, noise_type="gaussian")
        image_gauss = np.where(image == 0, zero_noise, image)

        final_segmentation = np.empty_like(image)
        initial_segmentation = seg_init

        for i in range(max_iter):
            if self.verbose:
                print(f">> Iteration: {i}")

            PLN = icm.prob_neighborhood(seg_init, beta, nclasses)
            PVE = com.prob_image(image_gauss, nclasses, mu, sigmasq, PLN)

            mu_upd, sigmasq_upd = com.update_param(image_gauss, PVE, mu, nclasses)
            ind = np.argsort(mu_upd)
            mu_upd = mu_upd[ind]
            sigmasq_upd = sigmasq_upd[ind]

            negll = com.negloglikelihood(image_gauss, mu_upd, sigmasq_upd, nclasses)
            final_segmentation, energy = icm.icm_ising(negll, beta, seg_init)

            energy_sum.append(energy[energy > -np.inf].sum())

            if self.save_history:
                self.segmentations.append(final_segmentation)
                self.pves.append(PVE)
                self.energies.append(energy)
                self.energies_sum.append(energy_sum[-1])

            if tolerance > 0 and i > 5:
                e_sum = np.asarray(energy_sum)
                tol = tolerance * (np.amax(e_sum) - np.amin(e_sum))

                e_end = e_sum[-5:]
                test_dist = np.abs(np.amax(e_end) - np.amin(e_end))

                if test_dist < tol:
                    break

            seg_init = final_segmentation
            mu = mu_upd
            sigmasq = sigmasq_upd

        PVE = PVE[..., 1:]

        return initial_segmentation, final_segmentation, PVE


def compute_directional_average(
    data,
    bvals,
    *,
    s0_map=None,
    masks=None,
    b0_mask=None,
    b0_threshold=50,
    low_signal_threshold=50,
):
    """
    Compute the mean signal for each unique b-value shell and fit a linear model.

    Parameters
    ----------
    data : ndarray
        The diffusion MRI data.
    bvals : ndarray
        The b-values corresponding to the diffusion data.
    s0_map : ndarray, optional
        Precomputed mean signal map for b=0 images.
    masks : ndarray, optional
        Precomputed masks for each unique b-value shell.
    b0_mask : ndarray, optional
        Precomputed mask for b=0 images.
    b0_threshold : float, optional
        The intensity threshold for a b=0 image.
    low_signal_threshold : float, optional
        The threshold below which a voxel is considered to have low signal.

    Returns
    -------
    P : float
        The slope of the linear model.
    V : float
        The intercept of the linear model.
    """
    if b0_mask is None:
        b0_mask = bvals < b0_threshold
    if masks is None:
        unique_bvals = np.unique(bvals)
        masks = bvals[:, np.newaxis] == unique_bvals[np.newaxis, 1:]
    if s0_map is None:
        s0_map = data[..., b0_mask].mean(axis=-1)

    if s0_map < low_signal_threshold:
        return 0, 0

    # Calculate the mean for each mask
    means = np.sum(data[:, np.newaxis] * masks, axis=0) / np.sum(masks, axis=0)

    # Normalize by s0, avoiding division by zero by adding 0.01 for stable division
    s_bvals = means / (s0_map[..., np.newaxis] + 0.01)

    # Avoid log(0) by adding 0.001 for stable linear regression fit
    s_bvals[s_bvals == 0] = 0.001
    s_log = y = np.log(s_bvals)

    xb = -np.log(np.arange(1, s_log.shape[-1] + 1))

    # Reshape xb for linear regression
    X = xb.reshape(-1, 1)

    # Fit linear model
    model = linear_model.LinearRegression()
    model.fit(X, y)
    P = model.coef_[0]
    V = model.intercept_

    return P, V


def dam_classifier(
    data, bvals, wm_threshold, *, b0_threshold=50, low_signal_threshold=50
):
    """Computes the P-map (fitting slope) on data to extract white and grey matter.

    See :footcite:p:`Cheng2020` for further details about the method.

    Parameters
    ----------
    data : ndarray
        The diffusion MRI data.
    bvals : ndarray
        The b-values corresponding to the diffusion data.
    wm_threshold : float
        The threshold below which a voxel is considered white matter.
    b0_threshold : float, optional
        The intensity threshold for a b=0 image.
    low_signal_threshold : float, optional
        The threshold below which a voxel is considered to have low signal.

    Returns
    -------
    wm_mask : ndarray
        A binary mask for white matter.
    gm_mask : ndarray
        A binary mask for grey matter.

    References
    ----------
    .. footbibliography::

    """
    # Precompute unique b-values, masks, and b=0 mask
    unique_bvals = np.unique(bvals)
    if len(unique_bvals) <= 2:
        raise ValueError("Insufficient unique b-values for fitting.")

    b0_mask = bvals < b0_threshold
    masks = bvals[:, np.newaxis] == unique_bvals[np.newaxis, 1:]

    # Precompute s0 (mean signal for b=0)
    s0_map = data[..., b0_mask].mean(axis=-1)

    # If the mean signal for b=0 is too low, set those voxels to 0 for both P and V
    valid_voxels = s0_map >= low_signal_threshold

    P_map = np.zeros(data.shape[:-1])
    for idx in range(data.shape[0] * data.shape[1] * data.shape[2]):
        i, j, k = np.unravel_index(idx, P_map.shape)
        if valid_voxels[i, j, k]:
            P, _ = compute_directional_average(
                data[i, j, k, :],
                bvals,
                masks=masks,
                b0_mask=b0_mask,
                s0_map=s0_map[i, j, k],
                low_signal_threshold=low_signal_threshold,
            )
            P_map[i, j, k] = P

    # Adding a small slope threshold for P_map to avoid 0 sloped background voxels
    wm_mask = (P_map <= wm_threshold) & (P_map > 0.01)
    # Grey matter has a higher P value than white matter
    gm_mask = P_map > wm_threshold

    return wm_mask, gm_mask



"""Various utilities common to nibabel release and maintenance tools.
"""
# Library imports
import os
import sys

# Useful shorthands
pjoin = os.path.join
cd = os.chdir

# Utility functions
def c(cmd):
    """Run system command, raise SystemExit if it returns an error."""
    print ("$",cmd)
    stat = os.system(cmd)
    #stat = 0  # Uncomment this and comment previous to run in debug mode
    if stat:
        raise SystemExit(f"Command {cmd} failed with code: {stat}")


def get_dipydir():
    """Get dipy directory from command line, or assume it's the one above."""

    # Initialize arguments and check location
    try:
        dipydir = sys.argv[1]
    except IndexError:
        dipydir = '..'

    dipydir = os.path.abspath(dipydir)

    cd(dipydir)
    if not os.path.isdir('dipy') and os.path.isfile('setup.py'):
        raise SystemExit(f'Invalid dipy directory: {dipydir}')
    return dipydir

# import compileall and then get dir os.path.split
def compile_tree():
    """Compile all Python files below current directory."""
    stat = os.system('python3 -m compileall .')
    if stat:
        msg = '*** ERROR: Some Python files in tree do NOT compile! ***\n'
        msg += 'See messages above for the actual file that produced it.\n'
        raise SystemExit(msg)



import nibabel as nib
import numpy as np
from numpy.testing import assert_, assert_array_almost_equal, assert_equal

import dipy.data as dpd
from dipy.denoise.adaptive_soft_matching import adaptive_soft_matching
from dipy.denoise.noise_estimate import estimate_sigma
from dipy.denoise.non_local_means import non_local_means
from dipy.testing.decorators import set_random_number_generator


def test_ascm_static():
    S0 = 100 * np.ones((20, 20, 20), dtype="f8")
    S0n1 = non_local_means(S0, sigma=0, rician=False, patch_radius=1, block_radius=1)
    S0n2 = non_local_means(S0, sigma=0, rician=False, patch_radius=2, block_radius=1)
    S0n = adaptive_soft_matching(S0, S0n1, S0n2, 0)
    assert_array_almost_equal(S0, S0n)


@set_random_number_generator()
def test_ascm_random_noise(rng):
    S0 = 100 + 2 * rng.standard_normal((22, 23, 30))
    S0n1 = non_local_means(S0, sigma=1, rician=False, patch_radius=1, block_radius=1)
    S0n2 = non_local_means(S0, sigma=1, rician=False, patch_radius=2, block_radius=1)
    S0n = adaptive_soft_matching(S0, S0n1, S0n2, 1)

    print(S0.mean(), S0.min(), S0.max())
    print(S0n.mean(), S0n.min(), S0n.max())

    assert_(S0n.min() > S0.min())
    assert_(S0n.max() < S0.max())
    assert_equal(np.round(S0n.mean()), 100)


@set_random_number_generator()
def test_ascm_rmse_with_nlmeans(rng):
    # checks the smoothness
    S0 = np.ones((30, 30, 30)) * 100
    S0[10:20, 10:20, 10:20] = 50
    S0[20:30, 20:30, 20:30] = 0
    S0_noise = S0 + 20 * rng.standard_normal((30, 30, 30))
    print("Original RMSE", np.sum(np.abs(S0 - S0_noise)) / np.sum(S0))

    S0n1 = non_local_means(
        S0_noise, sigma=400, rician=False, patch_radius=1, block_radius=1
    )
    print("Smaller patch RMSE", np.sum(np.abs(S0 - S0n1)) / np.sum(S0))
    S0n2 = non_local_means(
        S0_noise, sigma=400, rician=False, patch_radius=2, block_radius=2
    )
    print("Larger patch RMSE", np.sum(np.abs(S0 - S0n2)) / np.sum(S0))
    S0n = adaptive_soft_matching(S0, S0n1, S0n2, 400)
    print("ASCM RMSE", np.sum(np.abs(S0 - S0n)) / np.sum(S0))

    assert_(
        np.sum(np.abs(S0 - S0n)) / np.sum(S0) < np.sum(np.abs(S0 - S0n1)) / np.sum(S0)
    )
    assert_(
        np.sum(np.abs(S0 - S0n)) / np.sum(S0)
        < np.sum(np.abs(S0 - S0_noise)) / np.sum(S0)
    )
    assert_(90 < np.mean(S0n) < 110)


@set_random_number_generator()
def test_sharpness(rng):
    # check the edge-preserving nature
    S0 = np.ones((30, 30, 30)) * 100
    S0[10:20, 10:20, 10:20] = 50
    S0[20:30, 20:30, 20:30] = 0
    S0_noise = S0 + 20 * rng.standard_normal((30, 30, 30))
    S0n1 = non_local_means(
        S0_noise, sigma=400, rician=False, patch_radius=1, block_radius=1
    )
    edg1 = np.abs(np.mean(S0n1[8, 10:20, 10:20] - S0n1[12, 10:20, 10:20]) - 50)
    print("Edge gradient smaller patch", edg1)
    S0n2 = non_local_means(
        S0_noise, sigma=400, rician=False, patch_radius=2, block_radius=2
    )
    edg2 = np.abs(np.mean(S0n2[8, 10:20, 10:20] - S0n2[12, 10:20, 10:20]) - 50)
    print("Edge gradient larger patch", edg2)
    S0n = adaptive_soft_matching(S0, S0n1, S0n2, 400)
    edg = np.abs(np.mean(S0n[8, 10:20, 10:20] - S0n[12, 10:20, 10:20]) - 50)
    print("Edge gradient ASCM", edg)

    assert_(edg2 > edg1)
    assert_(edg2 > edg)
    assert_(np.abs(edg1 - edg) < 1.5)


def test_ascm_accuracy():
    f_name = dpd.get_fnames(name="ascm_test")
    test_ascm_data_ref = np.asanyarray(nib.load(f_name).dataobj)
    test_data = np.asanyarray(nib.load(dpd.get_fnames(name="aniso_vox")).dataobj)

    # the test data was constructed in this manner
    mask = test_data > 50
    sigma = estimate_sigma(test_data, N=4).item()

    den_small = non_local_means(
        test_data, sigma=sigma, mask=mask, patch_radius=1, block_radius=1, rician=True
    )

    den_large = non_local_means(
        test_data, sigma=sigma, mask=mask, patch_radius=2, block_radius=1, rician=True
    )

    S0n = np.array(adaptive_soft_matching(test_data, den_small, den_large, sigma))

    assert_array_almost_equal(S0n, test_ascm_data_ref, decimal=4)
